import os
import unittest

from parsons import Postgres, Table
from test.utils import assert_matching_tables

# The name of the schema and will be temporarily created for the tests
TEMP_SCHEMA = "parsons_test"

# These tests do not interact with the Postgres Database directly, and don't need real credentials


class TestPostgresCreateStatement(unittest.TestCase):
    def setUp(self):
        self.pg = Postgres(username="test", password="test", host="test", db="test", port=123)

        self.tbl = Table([["ID", "Name"], [1, "Jim"], [2, "John"], [3, "Sarah"]])

        self.tbl2 = Table(
            [
                ["c1", "c2", "c3", "c4", "c5", "c6", "c7"],
                ["a", "", 1, "NA", 1.4, 1, 2],
                ["b", "", 2, "NA", 1.4, 1, 2],
                ["c", "", 3.4, "NA", "", "", "a"],
                ["d", "", 5, "NA", 1.4, 1, 2],
                ["e", "", 6, "NA", 1.4, 1, 2],
                ["f", "", 7.8, "NA", 1.4, 1, 2],
                ["g", "", 9, "NA", 1.4, 1, 2],
            ]
        )
        self.mapping = self.pg.generate_data_types(self.tbl)
        self.mapping2 = self.pg.generate_data_types(self.tbl2)

    def test_connection(self):
        # Test connection with kwargs passed
        Postgres(username="test", password="test", host="test", db="test")

        # Test connection with env variables
        os.environ["PGUSER"] = "user_env"
        os.environ["PGPASSWORD"] = "pass_env"
        os.environ["PGHOST"] = "host_env"
        os.environ["PGDATABASE"] = "db_env"
        os.environ["PGPORT"] = "5432"
        pg_env = Postgres()

        self.assertEqual(pg_env.username, "user_env")
        self.assertEqual(pg_env.password, "pass_env")
        self.assertEqual(pg_env.host, "host_env")
        self.assertEqual(pg_env.db, "db_env")
        self.assertEqual(pg_env.port, 5432)

    def test_data_type(self):
        # Test smallint
        self.assertEqual(self.pg.data_type(1, ""), "smallint")
        self.assertEqual(self.pg.data_type(2, ""), "smallint")
        # Test int
        self.assertEqual(self.pg.data_type(32769, ""), "int")
        # Test bigint
        self.assertEqual(self.pg.data_type(2147483648, ""), "bigint")
        # Test varchar that looks like an int
        self.assertEqual(self.pg.data_type("00001", ""), "varchar")
        # Test a float as a decimal
        self.assertEqual(self.pg.data_type(5.001, ""), "decimal")
        # Test varchar
        self.assertEqual(self.pg.data_type("word", ""), "varchar")
        # Test int with underscore as string
        self.assertEqual(self.pg.data_type("1_2", ""), "varchar")
        # Test int with leading zero as string
        self.assertEqual(self.pg.data_type("01", ""), "varchar")
        # Test int with underscore
        self.assertEqual(self.pg.data_type(1_2, ""), "smallint")

        # Test bool
        self.assertEqual(self.pg.data_type(True, ""), "bool")

    def test_generate_data_types(self):
        # Test correct header labels
        self.assertEqual(self.mapping["headers"], ["ID", "Name"])
        # Test correct data types
        self.assertEqual(self.mapping["type_list"], ["smallint", "varchar"])
        self.assertEqual(
            self.mapping2["type_list"],
            [
                "varchar",
                "varchar",
                "decimal",
                "varchar",
                "decimal",
                "smallint",
                "varchar",
            ],
        )
        # Test correct lengths
        self.assertEqual(self.mapping["longest"], [1, 5])

    def test_vc_padding(self):
        # Test padding calculated correctly
        self.assertEqual(self.pg.vc_padding(self.mapping, 0.2), [1, 6])

    def test_vc_max(self):
        # Test max sets it to the max
        self.assertEqual(self.pg.vc_max(self.mapping, ["Name"]), [1, 65535])

        # Test raises when can't find column
        # To Do

    def test_vc_validate(self):
        # Test that a column with a width of 0 is set to 1
        self.mapping["longest"][0] = 0
        self.mapping = self.pg.vc_validate(self.mapping)
        self.assertEqual(self.mapping, [1, 5])

    def test_create_sql(self):
        # Test the the statement is expected
        sql = self.pg.create_sql("tmc.test", self.mapping, distkey="ID")
        exp_sql = "create table tmc.test (\n  id smallint,\n  name varchar(5)) \ndistkey(ID) ;"
        self.assertEqual(sql, exp_sql)

    def test_column_validate(self):
        bad_cols = [
            "a",
            "a",
            "",
            "SELECT",
            "asdfjkasjdfklasjdfklajskdfljaskldfjaklsdfjlaksdfjklasjdfklasjdkfljaskldfljkasjdkfasjlkdfjklasdfjklakjsfasjkdfljaslkdfjklasdfjklasjkldfakljsdfjalsdkfjklasjdfklasjdfklasdkljf",
        ]
        fixed_cols = [
            "a",
            "a_1",
            "col_2",
            "col_3",
            "asdfjkasjdfklasjdfklajskdfljaskldfjaklsdfjlaksdfjklasjdfklasjdkfljaskldfljkasjdkfasjlkdfjklasdfjklakjsfasjkdfljaslkdfjkl",
        ]
        self.assertEqual(self.pg.column_name_validate(bad_cols), fixed_cols)

    def test_create_statement(self):
        # Assert that copy statement is expected
        sql = self.pg.create_statement(self.tbl, "tmc.test", distkey="ID")
        exp_sql = (
            """create table tmc.test (\n  "id" smallint,\n  "name" varchar(5)) \ndistkey(ID) ;"""
        )
        self.assertEqual(sql, exp_sql)

        # Assert that an error is raised by an empty table
        empty_table = Table([["Col_1", "Col_2"]])
        self.assertRaises(ValueError, self.pg.create_statement, empty_table, "tmc.test")


# These tests interact directly with the Postgres database


@unittest.skipIf(not os.environ.get("LIVE_TEST"), "Skipping because not running live test")
class TestPostgresDB(unittest.TestCase):
    def setUp(self):
        self.temp_schema = TEMP_SCHEMA
        self.pg = Postgres()

        self.tbl = Table([["ID", "Name"], [1, "Jim"], [2, "John"], [3, "Sarah"]])

        # Create a schema, create a table, create a view
        setup_sql = f"""
                    drop schema if exists {self.temp_schema} cascade;
                    create schema {self.temp_schema};
                    """

        other_sql = f"""
                    create table {self.temp_schema}.test (id smallint,name varchar(5));
                    create view {self.temp_schema}.test_view as (select * from {self.temp_schema}.test);
                    """

        self.pg.query(setup_sql)

        self.pg.query(other_sql)

    def tearDown(self):
        # Drop the view, the table and the schema
        teardown_sql = f"""
                       drop schema if exists {self.temp_schema} cascade;
                       """
        self.pg.query(teardown_sql)

    def test_query(self):
        # Check that query sending back expected result
        r = self.pg.query("select 1")
        self.assertEqual(r[0]["?column?"], 1)

    def test_query_with_parameters(self):
        table_name = f"{self.temp_schema}.test"
        self.pg.copy(self.tbl, f"{self.temp_schema}.test", if_exists="append")

        sql = f"select * from {table_name} where name = %s"
        name = "Sarah"
        r = self.pg.query(sql, parameters=[name])
        self.assertEqual(r[0]["name"], name)

        sql = f"select * from {table_name} where name in (%s, %s)"
        names = ["Sarah", "John"]
        r = self.pg.query(sql, parameters=names)
        self.assertEqual(r.num_rows, 2)

    def test_copy(self):
        # Copy a table and ensure table exists
        self.pg.copy(self.tbl, f"{self.temp_schema}.test_copy", if_exists="drop")
        r = self.pg.query(f"select * from {self.temp_schema}.test_copy where name='Jim'")
        self.assertEqual(r[0]["id"], 1)

        # Copy table and ensure truncate works.
        self.pg.copy(self.tbl, f"{self.temp_schema}.test_copy", if_exists="truncate")
        tbl = self.pg.query(f"select count(*) from {self.temp_schema}.test_copy")
        self.assertEqual(tbl.first, 3)

        # Copy table and ensure that drop works.
        self.pg.copy(self.tbl, f"{self.temp_schema}.test_copy", if_exists="drop")
        tbl = self.pg.query(f"select count(*) from {self.temp_schema}.test_copy")
        self.assertEqual(tbl.first, 3)

        # Copy table and ensure that append works.
        self.pg.copy(self.tbl, f"{self.temp_schema}.test_copy", if_exists="append")
        tbl = self.pg.query(f"select count(*) from {self.temp_schema}.test_copy")
        self.assertEqual(tbl.first, 6)

        # Try to copy the table and ensure that default fail works.
        self.assertRaises(ValueError, self.pg.copy, self.tbl, f"{self.temp_schema}.test_copy")

        # Try to copy the table and ensure that explicit fail works.
        self.assertRaises(
            ValueError,
            self.pg.copy,
            self.tbl,
            f"{self.temp_schema}.test_copy",
            if_exists="fail",
        )

    def test_to_postgres(self):
        self.tbl.to_postgres(f"{self.temp_schema}.test_copy")
        r = self.pg.query(f"select * from {self.temp_schema}.test_copy where name='Jim'")
        self.assertEqual(r[0]["id"], 1)

    def test_from_postgres(self):
        tbl = Table([["id", "name"], [1, "Jim"], [2, "John"], [3, "Sarah"]])

        self.pg.copy(self.tbl, f"{self.temp_schema}.test_copy", if_exists="drop")
        out_tbl = self.tbl.from_postgres(f"SELECT * FROM {self.temp_schema}.test_copy")
        assert_matching_tables(out_tbl, tbl)


if __name__ == "__main__":
    unittest.main()



"""NGPVAN Code Endpoints"""

import logging

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class Codes(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def get_codes(self, name=None, supported_entities=None, parent_code_id=None, code_type=None):
        """
        Get codes.

        `Args:`
            name : str
                Filter by name of code.
            supported_entities: str
                Filter by supported entities.
            parent_code_id: str
                Filter by parent code id.
            code_type: str
                Filter by code type.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {
            "name": name,
            "supportedEntities": supported_entities,
            "parentCodeId": parent_code_id,
            "codeType": code_type,
            "$top": 200,
        }

        tbl = Table(self.connection.get_request("codes", params=params))
        logger.info(f"Found {tbl.num_rows} codes.")
        return tbl

    def get_code(self, code_id):
        """
        Get a code.

        `Args:`
            code_id : int
                The code id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        c = self.connection.get_request(f"codes/{code_id}")
        logger.debug(c)
        logger.info(f"Found code {code_id}.")
        return c

    def get_code_types(self):
        """
        Get code types.

        `Returns:`
            list
                A list of code types.
        """

        lst = self.connection.get_request("codeTypes")
        logger.info(f"Found {len(lst)} code types.")
        return lst

    def create_code(
        self,
        name=None,
        parent_code_id=None,
        description=None,
        code_type="SourceCode",
        supported_entities=None,
    ):
        """
        Create a code.

        `Args:`
            name: str
                The name of the code.
            parent_code_id: int
                A unique identifier for this code’s parent.
            description: str
                A description for this code, no longer than 200 characters.
            code_type: str
                The code type. ``Tag`` and ``SourceCode`` are valid values.
            supported_entities: list
                A list of dicts that enumerate the searchability and applicability rules of the
                code. You can find supported entities with the :meth:`code_supported_entities`

                .. highlight:: python
                .. code-block:: python

                    [
                        {
                         'name': 'Event',
                         'is_searchable': True,
                         'is_applicable': True
                        }
                        {
                         'name': 'Locations',
                         'start_time': '12-31-2018T13:00:00',
                         'end_time': '12-31-2018T14:00:00'
                        }
                    ]
        """

        json = {
            "parentCodeId": parent_code_id,
            "name": name,
            "codeType": code_type,
            "description": description,
        }

        if supported_entities:
            se = [
                {
                    "name": s["name"],
                    "isSearchable": s["is_searchable"],
                    "isApplicable": s["is_applicable"],
                }
                for s in supported_entities
            ]

            json["supportedEntities"] = se

        r = self.connection.post_request("codes", json=json)
        logger.info(f"Code {r} created.")
        return r

    def update_code(
        self,
        code_id,
        name=None,
        parent_code_id=None,
        description=None,
        code_type="SourceCode",
        supported_entities=None,
    ):
        """
        Update a code.

        `Args:`
            code_id: int
                The code id.
            name: str
                The name of the code.
            parent_code_id: int
                A unique identifier for this code’s parent.
            description: str
                A description for this code, no longer than 200 characters.
            code_type: str
                The code type. ``Tag`` and ``SourceCode`` are valid values.
            supported_entities: list
                A list of dicts that enumerate the searchability and applicability rules of the
                code. You can find supported entities with the :meth:`code_supported_entities`

                .. highlight:: python
                .. code-block:: python

                    [
                        {
                         'name': 'Event',
                         'is_searchable': True,
                         'is_applicable': True
                        }
                        {
                         'name': 'Locations',
                         'start_time': '12-31-2018T13:00:00',
                         'end_time': '12-31-2018T14:00:00'
                        }
                    ]
        """

        post_data = {}

        if name:
            post_data["name"] = name
        if parent_code_id:
            post_data["parentCodeId"] = parent_code_id
        if code_type:
            post_data["codeType"] = code_type
        if description:
            post_data["description"] = description

        if supported_entities:
            se = [
                {
                    "name": s["name"],
                    "isSearchable": s["is_searchable"],
                    "isApplicable": s["is_applicable"],
                }
                for s in supported_entities
            ]
            post_data["supportedEntities"] = se

        r = self.connection.put_request(f"codes/{code_id}", json=post_data)
        logger.info(f"Code {code_id} updated.")
        return r

    def delete_code(self, code_id):
        """
        Delete a code.

        `Args:`
            code_id: int
                The code id.
        `Returns:`
            ``None``
        """

        r = self.connection.delete_request(f"codes/{code_id}")
        logger.info(f"Code {code_id} deleted.")
        return r

    def get_code_supported_entities(self):
        """
        Get code supported entities.

        `Returns:`
            list
                A list of code supported entities.
        """

        lst = self.connection.get_request("codes/supportedEntities")
        logger.info(f"Found {len(lst)} code supported entities.")
        return lst



import os
import unittest
from unittest import mock

import requests_mock

from parsons import Sisense
from test.test_sisense.test_data import (
    ENV_PARAMETERS,
    TEST_DELETE_SHARED_DASHBOARD,
    TEST_LIST_SHARED_DASHBOARDS,
    TEST_PUBLISH_SHARED_DASHBOARD,
)


class TestSisense(unittest.TestCase):
    def setUp(self):
        self.sisense = Sisense(site_name="my_site_name", api_key="my_api_key")

    @mock.patch.dict(os.environ, ENV_PARAMETERS)
    def test_init(self):
        sisense = Sisense()
        self.assertEqual(sisense.site_name, "my_site_name")
        self.assertEqual(sisense.api_key, "my_api_key")
        self.assertEqual(sisense.api.uri, "https://app.periscopedata.com/api/v1/")
        self.assertEqual(sisense.api.headers["HTTP-X-PARTNER-AUTH"], "my_site_name:my_api_key")

    @requests_mock.Mocker()
    def test_publish_shared_dashboard(self, m):
        m.post(
            f"{self.sisense.uri}shared_dashboard/create",
            json=TEST_PUBLISH_SHARED_DASHBOARD,
        )
        self.assertEqual(
            self.sisense.publish_shared_dashboard(dashboard_id="1234"),
            TEST_PUBLISH_SHARED_DASHBOARD,
        )

    @requests_mock.Mocker()
    def test_list_shared_dashboards(self, m):
        m.post(f"{self.sisense.uri}shared_dashboard/list", json=TEST_LIST_SHARED_DASHBOARDS)
        self.assertEqual(
            self.sisense.list_shared_dashboards(dashboard_id="1234"),
            TEST_LIST_SHARED_DASHBOARDS,
        )

    @requests_mock.Mocker()
    def test_delete_shared_dashboard(self, m):
        m.post(
            f"{self.sisense.uri}shared_dashboard/delete",
            json=TEST_DELETE_SHARED_DASHBOARD,
        )
        self.assertEqual(
            self.sisense.delete_shared_dashboard(token="abc"),
            TEST_DELETE_SHARED_DASHBOARD,
        )



import os
import unittest
from unittest.mock import patch

import requests_mock
from github.GithubException import UnknownObjectException

from parsons import GitHub, Table
from parsons.github.github import ParsonsGitHubError

_dir = os.path.dirname(__file__)


class TestGitHub(unittest.TestCase):
    def setUp(self):
        self.github = GitHub(access_token="token")

    @requests_mock.Mocker()
    def test_wrap_github_404(self, m):
        with patch("github.Github.get_repo") as get_repo_mock:
            get_repo_mock.side_effect = UnknownObjectException("", "")
            with self.assertRaises(ParsonsGitHubError):
                self.github.get_repo("octocat/Hello-World")

    @requests_mock.Mocker()
    def test_get_repo(self, m):
        with open(os.path.join(_dir, "test_data", "test_get_repo.json"), "r") as f:
            m.get(requests_mock.ANY, text=f.read())
        repo = self.github.get_repo("octocat/Hello-World")
        self.assertEqual(repo["id"], 1296269)
        self.assertEqual(repo["name"], "Hello-World")

    @requests_mock.Mocker()
    def test_list_repo_issues(self, m):
        with open(os.path.join(_dir, "test_data", "test_get_repo.json"), "r") as f:
            m.get("https://api.github.com:443/repos/octocat/Hello-World", text=f.read())
        with open(os.path.join(_dir, "test_data", "test_list_repo_issues.json"), "r") as f:
            m.get(
                "https://api.github.com:443/repos/octocat/Hello-World/issues",
                text=f.read(),
            )
        issues_table = self.github.list_repo_issues("octocat/Hello-World")
        self.assertIsInstance(issues_table, Table)
        self.assertEqual(len(issues_table.table), 2)
        self.assertEqual(issues_table[0]["id"], 1)
        self.assertEqual(issues_table[0]["title"], "Found a bug")

    @requests_mock.Mocker()
    def test_download_file(self, m):
        with open(os.path.join(_dir, "test_data", "test_get_repo.json"), "r") as f:
            m.get("https://api.github.com:443/repos/octocat/Hello-World", text=f.read())
        with open(os.path.join(_dir, "test_data", "test_download_file.csv"), "r") as f:
            m.get(
                "https://raw.githubusercontent.com/octocat/Hello-World/testing/data.csv",
                text=f.read(),
            )

        file_path = self.github.download_file("octocat/Hello-World", "data.csv", branch="testing")
        with open(file_path, "r") as f:
            file_contents = f.read()

        self.assertEqual(file_contents, "header\ndata\n")



import base64
import io
import re
import unittest

from parsons import SMTP


class FakeConnection(object):
    def __init__(self, result_obj):
        self.result_obj = result_obj

    def sendmail(self, sender, to, message_body):
        self.result_obj.result = (sender, to, message_body)
        if "willfail@example.com" in to:
            return {"willfail@example.com": (550, "User unknown")}

    def quit(self):
        self.result_obj.quit_ran = True


class TestSMTP(unittest.TestCase):
    def setUp(self):
        self.smtp = SMTP("fake.example.com", username="fake", password="fake")
        self.smtp.conn = FakeConnection(self)
        self.result = None
        self.quit_ran = False

    def test_send_message_simple(self):
        self.smtp.send_email(
            "foo@example.com", "recipient1@example.com", "Simple subject", "Fake body"
        )
        self.assertEqual(self.result[0], "foo@example.com")
        self.assertEqual(self.result[1], ["recipient1@example.com"])
        self.assertTrue(
            self.result[2].endswith(
                "\nto: recipient1@example.com\nfrom: foo@example.com"
                "\nsubject: Simple subject\n\nFake body"
            )
        )
        self.assertTrue(self.quit_ran)

    def test_send_message_html(self):
        self.smtp.send_email(
            "foohtml@example.com",
            "recipienthtml@example.com",
            "Simple subject",
            "Fake body",
            "<p>Really Fake html</p>",
        )
        self.assertEqual(self.result[0], "foohtml@example.com")
        self.assertEqual(self.result[1], ["recipienthtml@example.com"])
        self.assertRegex(self.result[2], r"<p>Really Fake html</p>\n--=======")
        self.assertRegex(self.result[2], r"\nFake body\n--======")
        self.assertRegex(self.result[2], r"ubject: Simple subject\n")
        self.assertTrue(self.quit_ran)

    def test_send_message_manualclose(self):
        smtp = SMTP("fake.example.com", username="fake", password="fake", close_manually=True)
        smtp.conn = FakeConnection(self)
        smtp.send_email("foo@example.com", "recipient1@example.com", "Simple subject", "Fake body")
        self.assertFalse(self.quit_ran)

    def test_send_message_files(self):
        named_file_content = "x,y,z\n1,2,3\r\n3,4,5\r\n"
        unnamed_file_content = "foo,bar\n1,2\r\n3,4\r\n"
        bytes_file_content = bytes(
            [
                71,
                73,
                70,
                56,
                57,
                97,
                1,
                0,
                1,
                0,
                0,
                255,
                0,
                44,
                0,
                0,
                0,
                0,
                1,
                0,
                1,
                0,
                0,
                2,
                0,
                59,
            ]
        )
        named_file = io.StringIO(named_file_content)
        named_file.name = "xyz.csv"

        bytes_file = io.BytesIO(bytes_file_content)
        bytes_file.name = "xyz.gif"

        self.smtp.send_email(
            "foofiles@example.com",
            "recipientfiles@example.com",
            "Simple subject",
            "Fake body",
            files=[io.StringIO(unnamed_file_content), named_file, bytes_file],
        )
        self.assertEqual(self.result[0], "foofiles@example.com")
        self.assertEqual(self.result[1], ["recipientfiles@example.com"])
        self.assertRegex(self.result[2], r"\nFake body\n--======")
        found = re.findall(r'filename="file"\n\n([\w=/]+)\n\n--===', self.result[2])
        self.assertEqual(base64.b64decode(found[0]).decode(), unnamed_file_content)
        found_named = re.findall(
            r'Content-Type: text/csv; charset="utf-8"\nMIME-Version: 1.0'
            r"\nContent-Transfer-Encoding: base64\nContent-Disposition: "
            r'attachment; filename="xyz.csv"\n\n([\w=/]+)\n\n--======',
            self.result[2],
        )
        self.assertEqual(base64.b64decode(found_named[0]).decode(), named_file_content)

        found_gif = re.findall(
            r"Content-Type: image/gif\nMIME-Version: 1.0"
            r"\nContent-Transfer-Encoding: base64\nContent-ID: <xyz.gif>"
            r'\nContent-Disposition: attachment; filename="xyz.gif"\n\n([\w=/]+)\n\n--==',
            self.result[2],
        )
        self.assertEqual(base64.b64decode(found_gif[0]), bytes_file_content)
        self.assertTrue(self.quit_ran)

    def test_send_message_partial_fail(self):
        simple_msg = self.smtp._create_message_simple(
            "foo@example.com",
            "recipient1@example.com, willfail@example.com",
            "Simple subject",
            "Fake body",
        )
        send_result = self.smtp._send_message(simple_msg)
        self.assertEqual(send_result, {"willfail@example.com": (550, "User unknown")})



"""NGPVAN Target Endpoints"""

import logging

import petl

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class TargetsFailed(Exception):
    pass


class Targets(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def obj_dict(obj):
        return obj.__dict__

    def get_targets(self):
        """
        Get targets.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("targets"))
        logger.info(f"Found {tbl.num_rows} targets.")
        return tbl

    def get_target(self, target_id):
        """
        Get a single target.

        `Args:`
            target_id : int
                The target id.
        `Returns:`
            dict
                The target
        """

        r = self.connection.get_request(f"targets/{target_id}")
        logger.info(f"Found target {target_id}.")
        return r

    def get_target_export(self, export_job_id):
        """
        Get specific target export job id's status.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        response = self.connection.get_request(f"targetExportJobs/{export_job_id}")
        job_status = response.get("jobStatus")
        if job_status == "Complete":
            url = response["file"]["downloadUrl"]
            return Table(petl.fromcsv(url, encoding="utf-8-sig"))
        elif job_status == "Pending" or job_status == "InProcess":
            logger.info(f"Target export job is pending or in process for {export_job_id}.")
        else:
            raise TargetsFailed(f"Target export failed for {export_job_id}")

    def create_target_export(self, target_id, webhook_url=None):
        """
        Create new target export job

        `Args:`
            target_id : int
                The target id the export job is creating for.
        `Returns:`
            dict
                The target export job ID
        """
        target_export = {"targetId": target_id}

        r = self.connection.post_request("targetExportJobs", json=target_export)
        logger.info(f"Created new target export job for {target_id}.")
        return r



import os
import unittest

import requests_mock
from requests.exceptions import HTTPError

from parsons import VAN
from test.test_van.responses_people import (
    delete_person_response,
    find_people_response,
    get_person_response,
    merge_contacts_response,
)

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestNGPVAN(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    @requests_mock.Mocker()
    def test_find_person(self, m):
        m.post(
            self.van.connection.uri + "people/find",
            json=find_people_response,
            status_code=200,
        )

        person = self.van.find_person(first_name="Bob", last_name="Smith", phone=4142020792)

        self.assertEqual(person, find_people_response)

    @requests_mock.Mocker()
    def test_find_person_json(self, m):
        json = {
            "firstName": "Bob",
            "lastName": "Smith",
            "phones": [{"phoneNumber": 4142020792}],
        }

        m.post(
            self.van.connection.uri + "people/find",
            json=find_people_response,
            status_code=200,
        )

        person = self.van.find_person_json(match_json=json)

        self.assertEqual(person, find_people_response)

    def test_upsert_person(self):
        pass

    def test_upsert_person_json(self):
        pass

    def test_update_person(self):
        pass

    def test_update_person_json(self):
        pass

    def test_people_search(self):
        # Already tested as part of upsert and find person methods
        pass

    def test_valid_search(self):
        # Fails with FN / LN Only
        self.assertRaises(
            ValueError,
            self.van._valid_search,
            "Barack",
            "Obama",
            None,
            None,
            None,
            None,
            None,
        )

        # Fails with only Zip
        self.assertRaises(
            ValueError,
            self.van._valid_search,
            "Barack",
            "Obama",
            None,
            None,
            None,
            None,
            60622,
        )

        # Fails with no street number
        self.assertRaises(
            ValueError,
            self.van._valid_search,
            "Barack",
            "Obama",
            None,
            None,
            None,
            "Pennsylvania Ave",
            None,
        )

        # Successful with FN/LN/Email
        self.van._valid_search("Barack", "Obama", "barack@email.com", None, None, None, None)

        # Successful with FN/LN/DOB/ZIP
        self.van._valid_search(
            "Barack", "Obama", "barack@email.com", None, "2000-01-01", None, 20009
        )

        # Successful with FN/LN/Phone
        self.van._valid_search("Barack", "Obama", None, 2024291000, None, None, None)

    @requests_mock.Mocker()
    def test_get_person(self, m):
        json = get_person_response

        # Test works with external ID
        m.get(self.van.connection.uri + "people/DWID:15406767", json=json)
        person = self.van.get_person("15406767", id_type="DWID")
        self.assertEqual(get_person_response, person)

        # Test works with vanid
        m.get(self.van.connection.uri + "people/19722445", json=json)
        person = self.van.get_person("19722445")
        self.assertEqual(get_person_response, person)

    @requests_mock.Mocker()
    def test_delete_person(self, m):
        json = delete_person_response
        # Test works with vanid
        m.delete(self.van.connection.uri + "people/19722445", json=json)
        response = self.van.delete_person("19722445")
        self.assertEqual(delete_person_response, response)

    @requests_mock.Mocker()
    def test_apply_canvass_result(self, m):
        # Test a valid attempt
        m.post(self.van.connection.uri + "people/2335282/canvassResponses", status_code=204)
        self.van.apply_canvass_result(2335282, 18)

        # Test a bad result code
        json = {
            "errors": [
                {
                    "code": "INVALID_PARAMETER",
                    "text": "'resultCodeId' must be a valid result code in the current context.",
                    "properties": ["resultCodeId"],
                }
            ]
        }
        m.post(
            self.van.connection.uri + "people/2335282/canvassResponses",
            json=json,
            status_code=400,
        )
        self.assertRaises(HTTPError, self.van.apply_canvass_result, 2335282, 0)

        # Test a bad vanid
        json = {
            "errors": [
                {
                    "code": "INTERNAL_SERVER_ERROR",
                    "text": "An unknown error occurred",
                    "referenceCode": "88A111-E2FF8",
                }
            ]
        }
        m.post(
            self.van.connection.uri + "people/0/canvassResponses",
            json=json,
            status_code=400,
        )
        self.assertRaises(HTTPError, self.van.apply_canvass_result, 0, 18)

        # Test a good dwid
        m.post(
            self.van.connection.uri + "people/DWID:2335282/canvassResponses",
            status_code=204,
        )
        self.van.apply_canvass_result(2335282, 18, id_type="DWID")

        # test canvassing via phone or sms without providing phone number
        self.assertRaises(Exception, self.van.apply_canvass_result, 2335282, 18, contact_type_id=37)

        # test canvassing via phone or sms with providing phone number
        m.post(self.van.connection.uri + "people/2335282/canvassResponses", status_code=204)
        self.van.apply_canvass_result(2335282, 18, contact_type_id=37, phone="(516)-555-2342")

    @requests_mock.Mocker()
    def test_apply_survey_question(self, m):
        # Test valid survey question
        m.post(self.van.connection.uri + "people/2335282/canvassResponses", status_code=204)
        self.van.apply_survey_response(2335282, 351006, 1443891)

        # Test bad survey response id
        # json = {
        #     'errors': [{
        #         'code': 'INVALID_PARAMETER',
        #         'text': ("'surveyResponseId' must be a valid Response to the given "
        #                  "Survey Question."),
        #         'properties': ['responses[0].surveyResponseId']
        #     }]
        # }
        m.post(self.van.connection.uri + "people/2335282/canvassResponses", status_code=400)
        self.assertRaises(HTTPError, self.van.apply_survey_response, 2335282, 0, 1443891)

        # Test bad survey question id
        # json = {
        #     'errors': [{
        #         'code': 'INVALID_PARAMETER',
        #         'text': ("'surveyQuestionId' must be a valid Survey Question that is "
        #                 "available in the current context."),
        #         'properties': ['responses[0].surveyQuestionId']
        #     }]
        # }
        m.post(self.van.connection.uri + "people/2335282/canvassResponses", status_code=400)
        self.assertRaises(HTTPError, self.van.apply_survey_response, 2335282, 351006, 0)

    def test_toggle_volunteer_action(self):
        pass

    def test_apply_response(self):
        pass

    @requests_mock.Mocker()
    def test_create_relationship(self, m):
        relationship_id = 12
        bad_vanid_1 = 99999
        good_vanid_1 = 12345
        vanid_2 = 54321

        # Bad request
        m.post(
            self.van.connection.uri + "people/{}/relationships".format(bad_vanid_1),
            status_code=404,
        )

        # Good request
        m.post(
            self.van.connection.uri + "people/{}/relationships".format(good_vanid_1),
            status_code=204,
        )

        # Test bad input
        self.assertRaises(
            HTTPError,
            self.van.create_relationship,
            bad_vanid_1,
            vanid_2,
            relationship_id,
        )
        self.assertRaises(
            HTTPError,
            self.van.create_relationship,
            bad_vanid_1,
            vanid_2,
            relationship_id,
        )

        self.van.create_relationship(good_vanid_1, vanid_2, relationship_id)

    @requests_mock.Mocker()
    def test_apply_person_code(self, m):
        vanid = 999
        code_id = 888

        # Test good request
        m.post(self.van.connection.uri + f"people/{vanid}/codes", status_code=204)
        self.van.apply_person_code(vanid, code_id)

        # Test bad request
        m.post(self.van.connection.uri + f"people/{vanid}/codes", status_code=404)
        self.assertRaises(HTTPError, self.van.apply_person_code, vanid, code_id)

    @requests_mock.Mocker()
    def test_merge_contacts(self, m):
        source_vanid = 12345

        m.put(
            self.van.connection.uri + f"people/{source_vanid}/mergeInto",
            json=merge_contacts_response,
            status_code=200,
        )

        person = self.van.merge_contacts(source_vanid=source_vanid, primary_vanid=56789)

        self.assertEqual(person, merge_contacts_response)



import os
import unittest
import unittest.mock as mock

import requests_mock

from parsons import VAN, Table
from parsons.utilities import cloud_storage
from test.utils import assert_matching_tables

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestBulkImport(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    @requests_mock.Mocker()
    def test_get_bulk_import_resources(self, m):
        json = ["Contacts", "Contributions", "ActivistCodes", "ContactsActivistCodes"]

        m.get(self.van.connection.uri + "bulkImportJobs/resources", json=json)

        self.assertEqual(self.van.get_bulk_import_resources(), json)

    @requests_mock.Mocker()
    def test_get_bulk_import_job(self, m):
        m.get(self.van.connection.uri + "bulkImportJobs/53407", json=bulk_import_job)

        self.assertEqual(self.van.get_bulk_import_job(53407), bulk_import_job)

    @requests_mock.Mocker()
    def test_get_bulk_import_job_results(self, m):
        results_tbl = Table(
            [
                [
                    "BulkUploadDataID",
                    "ULFileID",
                    "PrimaryKey",
                    "PrimaryKeyType",
                    "MailingAddress_3581",
                ],
                ["1", "1983", "101596008", "VanID", "Processed"],
            ]
        )

        bulk_import_job = {
            "id": 92,
            "status": "Completed",
            "resourceType": "Contacts",
            "webhookUrl": None,
            "resultFileSizeLimitKb": 5000,
            "errors": [],
            "resultFiles": [
                {
                    "url": Table.to_csv(results_tbl),
                    "dateExpired": "2020-09-04T22:07:04.0770295-04:00",
                }
            ],
        }

        m.get(self.van.connection.uri + "bulkImportJobs/53407", json=bulk_import_job)
        assert_matching_tables(self.van.get_bulk_import_job_results(53407), results_tbl)

    @requests_mock.Mocker()
    def test_get_bulk_import_mapping_types(self, m):
        m.get(self.van.connection.uri + "bulkImportMappingTypes", json=[mapping_type])

        assert_matching_tables(self.van.get_bulk_import_mapping_types(), Table([mapping_type]))

    @requests_mock.Mocker()
    def test_get_bulk_import_mapping_type(self, m):
        m.get(
            self.van.connection.uri + "bulkImportMappingTypes/ActivistCode",
            json=mapping_type,
        )

        self.assertEqual(self.van.get_bulk_import_mapping_type("ActivistCode"), mapping_type)

    @requests_mock.Mocker()
    def get_bulk_import_mapping_type_fields(self, m):
        json = [
            {"name": "Unsubscribed", "id": "0", "parents": None},
            {"name": "Not Subscribed", "id": "1", "parents": None},
            {"name": "Subscribed", "id": "2", "parents": None},
        ]
        m.get(
            self.van.connection.uri
            + "bulkImportMappingTypes/Email/EmailSubscriptionStatusId/values"
        )

        r = self.van.get_bulk_import_mapping_type_fields("Email", "EmailSubscriptionStatusId")
        self.assertEqual(json, r)

    @requests_mock.Mocker()
    def test_post_bulk_import(self, m):
        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://s3.com/my_file.zip"

        tbl = Table([["Vanid", "ActivistCodeID"], [1234, 345345]])

        m.post(self.van.connection.uri + "bulkImportJobs", json={"jobId": 54679})

        r = self.van.post_bulk_import(
            tbl,
            "S3",
            "ContactsActivistCodes",
            [{"name": "ActivistCode"}],
            "Activist Code Upload",
            bucket="my-bucket",
        )

        self.assertEqual(r, 54679)

    @requests_mock.Mocker()
    def test_bulk_apply_activist_codes(self, m):
        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://s3.com/my_file.zip"

        tbl = Table([["Vanid", "ActivistCodeID"], [1234, 345345]])

        m.post(self.van.connection.uri + "bulkImportJobs", json={"jobId": 54679})

        job_id = self.van.bulk_apply_activist_codes(tbl, url_type="S3", bucket="my-bucket")

        self.assertEqual(job_id, 54679)

    @requests_mock.Mocker()
    def test_bulk_apply_suppressions(self, m):
        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://s3.com/my_file.zip"

        tbl = Table([["Vanid", "suppressionid"], [1234, 18]])

        m.post(self.van.connection.uri + "bulkImportJobs", json={"jobId": 54679})

        job_id = self.van.bulk_apply_suppressions(tbl, url_type="S3", bucket="my-bucket")

        self.assertEqual(job_id, 54679)

    @requests_mock.Mocker()
    def test_bulk_upsert_contacts(self, m):
        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://s3.com/my_file.zip"

        tbl = Table([["Vanid", "email"], [1234, "me@me.com"]])

        m.post(self.van.connection.uri + "bulkImportJobs", json={"jobId": 54679})

        job_id = self.van.bulk_upsert_contacts(tbl, url_type="S3", bucket="my-bucket")

        self.assertEqual(job_id, 54679)

    @requests_mock.Mocker()
    def test_bulk_apply_canvass_results(self, m):
        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://s3.com/my_file.zip"

        tbl = Table(
            [
                ["vanid", "contacttypeid", "resultid", "datecanvassed", "canvassedby", "phone"],
                [1234, 1, 1, "2020-01-01", 987, "5554443210"],
            ]
        )

        m.post(self.van.connection.uri + "bulkImportJobs", json={"jobId": 54679})

        job_id = self.van.bulk_apply_canvass_results(tbl, url_type="S3", bucket="my-bucket")

        self.assertEqual(job_id, 54679)

    @requests_mock.Mocker()
    def test_bulk_apply_contact_custom_fields(self, m):
        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://s3.com/my_file.zip"

        tbl = Table([["vanid", "CF123", "CF124"], [1234, "Test String Value", 999]])

        m.post(self.van.connection.uri + "bulkImportJobs", json={"jobId": 54679})

        custom_field_group_id = 1234

        job_id = self.van.bulk_apply_contact_custom_fields(
            custom_field_group_id, tbl, url_type="S3", bucket="my-bucket"
        )

        self.assertEqual(job_id, 54679)


mapping_type = {
    "name": "ActivistCode",
    "displayName": "Apply Activist Code",
    "allowMultipleMode": "Multiple",
    "resourceTypes": ["Contacts", "ContactsActivistCodes"],
    "fields": [
        {
            "name": "ActivistCodeID",
            "description": "Activist Code ID",
            "hasPredefinedValues": True,
            "isRequired": True,
            "canBeMappedToColumn": True,
            "canBeMappedByName": True,
            "parents": None,
        },
        {
            "name": "CanvassedBy",
            "description": "Recruited By, Must be a Valid User ID",
            "hasPredefinedValues": False,
            "isRequired": False,
            "canBeMappedToColumn": True,
            "canBeMappedByName": True,
            "parents": None,
        },
        {
            "name": "DateCanvassed",
            "description": "Contacted When",
            "hasPredefinedValues": False,
            "isRequired": False,
            "canBeMappedToColumn": True,
            "canBeMappedByName": True,
            "parents": [{"parentFieldName": "CanvassedBy", "limitedToParentValues": None}],
        },
        {
            "name": "ContactTypeID",
            "description": "Contacted How",
            "hasPredefinedValues": True,
            "isRequired": False,
            "canBeMappedToColumn": True,
            "canBeMappedByName": True,
            "parents": [{"parentFieldName": "CanvassedBy", "limitedToParentValues": None}],
        },
    ],
}

bulk_import_job = {
    "id": 92,
    "status": "Completed",
    "resourceType": "Contacts",
    "webhookUrl": None,
    "resultFileSizeLimitKb": 5000,
    "errors": [],
    "resultFiles": [
        {
            "url": "https://ngpvan.com/bulk-import-jobs/f023.csv",
            "dateExpired": "2020-09-04T22:07:04.0770295-04:00",
        }
    ],
}



import logging

logger = logging.getLogger(__name__)


class Introspection(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def get_apikeyprofiles(self):
        """
        Get API key profiles.

        `Args:`
            None
        `Returns:`
            JSON response
        """

        response = self.connection.get_request("apiKeyProfiles")
        logger.info(f"Returned {len(response[0])} API key profiles.")
        return response[0]



import datetime
import logging
import pickle
import shutil
import sqlite3
import subprocess
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Literal, Optional, Union

import petl

from parsons.databases.database_connector import DatabaseConnector
from parsons.databases.table import BaseTable
from parsons.etl.table import Table
from parsons.utilities import files

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000

logger = logging.getLogger(__name__)


class SqliteTable(BaseTable):
    sql_placeholder = "?"

    def truncate(self) -> None:
        """
        Truncate the table.
        """
        self.db.query(f"delete from {self.table}")
        logger.info(f"{self.table} truncated.")


class Sqlite(DatabaseConnector):
    def __init__(self, db_path):
        self.db_path = db_path

    @contextmanager
    def connection(self) -> Iterator[sqlite3.Connection]:
        conn = sqlite3.connect(self.db_path)
        try:
            yield conn
        except sqlite3.Error:
            conn.rollback()
            raise
        else:
            conn.commit()
        finally:
            conn.close()

    @contextmanager
    def cursor(self, connection) -> Iterator[sqlite3.Cursor]:
        cur = connection.cursor()

        try:
            yield cur
        finally:
            cur.close()

    def query(self, sql: str, parameters: Optional[Union[list, dict]] = None) -> Optional[Table]:
        with self.connection() as connection:
            return self.query_with_connection(sql, connection, parameters=parameters)

    def query_with_connection(
        self,
        sql: str,
        connection: sqlite3.Connection,
        parameters: Optional[Union[list, dict]] = None,
        commit: bool = True,
        return_values: bool = True,
    ):
        """
        Execute a query against the database, with an existing connection. Useful for batching
        queries together. Will return ``None`` if the query returns zero rows.

        `Args:`
            sql: str
                A valid SQL statement
            connection: obj
                A connection object obtained from ``redshift.connection()``
            parameters: list
                A list of python variables to be converted into SQL values in your query
            commit: boolean
                Whether to commit the transaction immediately. If ``False`` the transaction will
                be committed when the connection goes out of scope and is closed (or you can
                commit manually with ``connection.commit()``).

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        # sqlite3 cursor cannot take None for parameters
        if not parameters:
            parameters = ()

        with self.cursor(connection) as cursor:
            logger.debug(f"SQL Query: {sql}")
            cursor.execute(sql, parameters)

            if commit:
                connection.commit()

            # Fetch the data in batches, and "pickle" the rows to a temp file.
            # (We pickle rather than writing to, say, a CSV, so that we maintain
            # all the type information for each field.)

            if return_values and cursor.description:
                temp_file = files.create_temp_file()

                with open(temp_file, "wb") as f:
                    # Grab the header
                    header = [i[0] for i in cursor.description]
                    pickle.dump(header, f)

                    while True:
                        batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                        if not batch:
                            break

                        logger.debug(f"Fetched {len(batch)} rows.")
                        for row in batch:
                            pickle.dump(list(row), f)

                # Load a Table from the file
                final_tbl = Table(petl.frompickle(temp_file))

                logger.debug(f"Query returned {final_tbl.num_rows} rows.")
                return final_tbl

    def generate_data_types(self, table: Table) -> dict[str, str]:
        """Generate column data types"""
        records = petl.records(table.table)

        type_list: dict[str, str] = {
            column: self._best_type(records, column) for column in table.columns
        }

        return type_list

    def _best_type(
        self, records: list[dict], column: str
    ) -> Literal["text", "integer", "float", "datetime", "date"]:
        values = [record[column] for record in records if record[column]]
        not_nulls = [i for i in values if i is not None]
        if not_nulls:
            not_null_value = not_nulls[0]
        else:
            not_null_value = None

        if isinstance(not_null_value, int) or isinstance(not_null_value, bool):
            result = "integer"
        elif isinstance(not_null_value, float):
            result = "float"
        elif isinstance(not_null_value, datetime.date):
            result = "date"
        elif isinstance(not_null_value, datetime.datetime):
            result = "datetime"
        else:
            result = "text"

        return result

    def create_statement(
        self,
        tbl,
        table_name,
    ) -> str:
        if tbl.num_rows == 0:
            raise ValueError("Table is empty. Must have 1 or more rows.")

        data_types = self.generate_data_types(tbl)
        result = "CREATE TABLE {} ({})".format(
            table_name,
            ", ".join(f"{column} {type}" for column, type in data_types.items()),
        )
        return result

    def copy(
        self,
        tbl: Table,
        table_name: str,
        if_exists: str = "fail",
        strict_length: bool = False,
        force_python_sdk: bool = False,
    ):
        """
        Copy a :ref:`parsons-table` to Sqlite.

        `Args:`
            tbl: parsons.Table
                A Parsons table object
            table_name: str
                The destination schema and table (e.g. ``my_schema.my_table``)
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            strict_length: bool
                If the database table needs to be created, strict_length determines whether
                the created table's column sizes will be sized to exactly fit the current data,
                or if their size will be rounded up to account for future values being larger
                then the current dataset. Defaults to ``False``.
            force_python_sdk: bool
                Use the python SDK to import data to sqlite3, even if the sqlite3 cli utility is available for more efficient loading. Defaults to False.
        """

        with self.connection() as connection:
            # Auto-generate table
            if self._create_table_precheck(connection, table_name, if_exists):
                # Create the table
                sql = self.create_statement(tbl, table_name)

                self.query_with_connection(sql, connection, commit=False, return_values=False)
                logger.info(f"{table_name} created.")

        # Use the sqlite3 command line for csv import if possible, as it is much more efficient
        if shutil.which("sqlite3") and not force_python_sdk:
            csv_file_path = tbl.to_csv()
            self._cli_command(f".import --csv --skip 1 {csv_file_path} {table_name}")
        else:
            self.import_table_iteratively(tbl, table_name, if_exists)

        logger.info(f"{len(tbl)} rows copied to {table_name}.")

    def import_table_iteratively(
        self, tbl: Table, table_name: str, if_exists: str, chunksize=10000
    ) -> None:
        """Import a CSV row by row using the python sqlite3 API.

        Iterates over chunks of length `chunksize`

        It is generally more efficient to use the sqlite3 CLI to
        import a CSV, but not all machines have the shell utility
        available, so we can fall back to this method.
        """
        chunked_tbls = tbl.chunk(chunksize)
        insert_sql = "INSERT INTO {} ({}) VALUES ({});".format(
            table_name,
            ", ".join(tbl.columns),
            ", ".join(["?" for _ in tbl.columns]),
        )
        with self.connection() as connection:
            with self.cursor(connection) as cursor:
                for chunked_tbl in chunked_tbls:
                    cursor.executemany(
                        insert_sql,
                        tuple([tuple(row.values()) for row in chunked_tbl]),
                    )

    def _cli_command(self, command: str) -> None:
        """Use the sqlite3 command line utility to run a command.

        Certain commands are only possible via the shell utility and
        not via the python API, such as the CSV import command.

        sqlite3 comes as part of the python stdlib, but the shell
        utility is not available by default on all systems. Windows
        machines in particular generally don't have the sqlite3
        utility unless it is explicitly installed.
        """
        db_path = Path(self.db_path).resolve()
        full_command = ["sqlite3", str(db_path), command]
        resp = subprocess.run(
            full_command,
            capture_output=True,
        )
        if resp.stderr:
            raise RuntimeError(resp.stderr.decode())
        if resp.returncode:
            raise RuntimeError(resp.stdout.decode())

    def _create_table_precheck(self, connection, table_name, if_exists) -> bool:
        """
        Helper to determine what to do when you need a table that may already exist.

        `Args:`
            connection: obj
                A connection object obtained from ``redshift.connection()``
            table_name: str
                The table to check
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``,
                or ``truncate`` the table.
        `Returns:`
            bool
                True if the table needs to be created, False otherwise.
        """

        if if_exists not in ["fail", "truncate", "append", "drop"]:
            raise ValueError("Invalid value for `if_exists` argument")

        # If the table exists, evaluate the if_exists argument for next steps.
        if self.table_exists(table_name):
            if if_exists == "fail":
                raise ValueError("Table already exists.")

            if if_exists == "truncate":
                truncate_sql = f"DELETE FROM {table_name};"
                logger.info(f"Truncating {table_name}.")
                self.query_with_connection(truncate_sql, connection, commit=False)

            if if_exists == "drop":
                logger.info(f"Dropping {table_name}.")
                drop_sql = f"DROP TABLE {table_name};"
                self.query_with_connection(drop_sql, connection, commit=False)
                return True

            return False

        else:
            return True

    def table_exists(self, table_name: str, view: bool = False) -> bool:
        """
        Check if a table or view exists in the database.

        `Args:`
            table_name: str
                The table name and schema (e.g. ``myschema.mytable``).
            view: boolean
                Check to see if a view exists by the same name. Defaults to ``False``.

        `Returns:`
            boolean
                ``True`` if the table exists and ``False`` if it does not.
        """
        # Check in pg tables for the table
        sql = "select name from sqlite_master where type=:type and name = :name"
        result = self.query(
            sql, parameters={"type": "view" if view else "table", "name": table_name}
        )

        # If in either, return boolean
        if result:
            return True
        else:
            return False

    def table(self, table_name: str) -> SqliteTable:
        # Return a Sqlite table instance
        return SqliteTable(self, table_name)



"""Logging classes for use with Parsons dbt utility."""

import datetime
import logging
import time
import uuid
from abc import ABC, abstractmethod
from typing import Optional

from dbt.contracts.graph.manifest import Manifest
from rich.console import Console
from rich.logging import RichHandler
from rich.markdown import Markdown

from parsons import Table
from parsons.databases.database_connector import DatabaseConnector
from parsons.utilities.dbt.models import EnhancedNodeResult

logger = logging.getLogger(__name__)


def human_readable_duration(seconds: int | float) -> str:
    time_struct = time.gmtime(seconds)

    time_str = ""

    if time_struct.tm_yday - 1:
        time_str += f"{time_struct.tm_yday - 1} days, "
    if time_struct.tm_hour:
        time_str += f"{time_struct.tm_hour} hours, "
    if time_struct.tm_min:
        time_str += f"{time_struct.tm_min} minutes, "
    if time_struct.tm_sec:
        time_str += f"{time_struct.tm_sec} seconds"

    return time_str


class dbtLogger(ABC):
    """Abstract base class for aggregating logs from dbt commands."""

    commands: list[Manifest]

    @abstractmethod
    def format_command_result(self, manifest: Manifest) -> str:
        pass

    @abstractmethod
    def format_result(self) -> str:
        pass

    @abstractmethod
    def send(self, manifests: list[Manifest]) -> None:
        """The send method is called to execute logging.

        manifests are passed to this method directly (rather
        than on initialization) so that the logger class can be
        initialized before the dbt commands have been run. This is
        mostly necessary for loggers that need to be initialized with
        credentials or options before being provided to the
        run_dbt_commands method.
        """
        self.commands = manifests
        log_text = self.format_result()  # noqa
        ...


class dbtLoggerMarkdown(dbtLogger):
    def format_command_result(
        self,
        manifest: Manifest,
    ) -> str:
        log_message = ""

        # Header
        if manifest.errors:
            log_message += "\U0001f534"  # Red box
            status = "Error"
        elif manifest.warnings:
            log_message += "\U0001f7e0"  # Orange circle
            status = "Warning"
        else:
            log_message += "\U0001f7e2"  # Green circle
            status = "Success"

        time_str = human_readable_duration(manifest.elapsed_time)
        log_message += f"Invoke dbt with `dbt {manifest.command}` ({status} in {time_str})"

        log_summary_str = ", ".join(
            [f"{node}: {count}" for node, count in manifest.summary.items()]
        )
        if not log_summary_str:
            log_summary_str = "No models ran."
        log_message += "\n*Summary*: `{}`".format(log_summary_str)

        log_message += "\n*GB Processed*: {:.2f}".format(manifest.total_gb_processed)
        log_message += "\n*Slot hours*: {:.2f}".format(manifest.total_slot_hours)

        # Errors
        if manifest.errors or manifest.fails:
            log_message += "\nError messages:\n```{}```".format(
                "\n\n".join(
                    [
                        i.node.name + ": " + (EnhancedNodeResult.log_message(i) or "")
                        for i in [*manifest.errors, *manifest.fails]
                    ]
                )
            )

        # Warnings
        if manifest.warnings:
            log_message += "\nWarn messages:\n```{}```".format(
                "\n\n".join(
                    [
                        i.node.name + ": " + (EnhancedNodeResult.log_message(i) or "")
                        for i in manifest.warnings
                    ]
                )
            )

        # Skips
        if manifest.skips:
            skips = {i.node.name for i in manifest.skips}
            log_message += "\nSkipped:\n```{}```".format(", ".join(skips))

        return log_message

    def format_result(self) -> str:
        """Format result string from all commands."""
        full_log_message = ""

        # Header
        if any(command.errors for command in self.commands):
            status = "failed"
            full_log_message += "\U0001f534"
        else:
            status = "succeeded"
            full_log_message += "\U0001f7e2"

        now = datetime.datetime.today().strftime("%Y-%m-%d %H:%M")
        full_log_message += f"*dbt run {status} - {now}*"

        total_duration = sum([command.elapsed_time for command in self.commands])
        duration_time_str = human_readable_duration(total_duration)
        full_log_message += f"\n*Duration:* {duration_time_str}\n\n"

        # Formatted results from each command
        log_messages = [self.format_command_result(command) for command in self.commands]
        full_log_message += "\n".join(log_messages)

        return full_log_message


class dbtLoggerStdout(dbtLoggerMarkdown):
    def send(self, manifests: list[Manifest]) -> None:
        self.commands = manifests
        log_text = self.format_result()

        md = Markdown(log_text)
        console = Console()
        console.print(md)


class dbtLoggerPython(dbtLoggerMarkdown):
    def send(self, manifests: list[Manifest]) -> None:
        self.commands = manifests
        log_text = self.format_result()

        if "RichHandler" not in [handler.__class__.__name__ for handler in logger.handlers]:
            logger.addHandler(RichHandler(level=logging.INFO))

        logger.info(log_text)


class dbtLoggerSlack(dbtLoggerMarkdown):
    def __init__(
        self,
        slack_webhook: str,
        slack_channel: Optional[str] = None,
    ) -> None:
        self.slack_webhook = slack_webhook
        self.slack_channel = slack_channel

    def send(self, manifests: list[Manifest]) -> None:
        self.commands = manifests
        log_text = self.format_result()

        # Importing here to avoid needing to make slackclient a dependency for all dbt users
        from parsons.notifications.slack import Slack

        Slack.message(channel=self.slack_channel, text=log_text, webhook=self.slack_webhook)


class dbtLoggerDatabase(dbtLogger, ABC):
    """Log dbt artifacts by loading to a database.

    This class is an abstract base class for logging dbt artifacts to
    a database.
    """

    def __init__(
        self,
        database_connector: DatabaseConnector,
        destination_table_runs: str,
        destination_table_nodes: str,
        extra_run_table_fields: dict,
        **copy_kwargs,
    ) -> None:
        """Initialize the logger.

        Args:
            database_connector: A DatabaseConnector object.
            destination_table_runs: The name of the table to log run information.
            destination_table_nodes: The name of the table to log node information.
            extra_run_table_fields: A dictionary of additional fields to include in the run table.
            **copy_kwargs: Additional keyword arguments to pass to the `copy` method.
        """
        self.db_connector = database_connector
        self.destination_table_runs = destination_table_runs
        self.destination_table_nodes = destination_table_nodes
        self.extra_run_table_fields = extra_run_table_fields
        self.copy_kwargs = copy_kwargs

    def format_command_result(self, manifest: Manifest) -> tuple[Table, Table]:
        """Loads all artifact results into a Parsons Table."""
        dbt_run_id = str(uuid.uuid4())
        run_metadata = {
            key: getattr(manifest, key) for key in ("command", "generated_at", "elapsed_time")
        }
        run_metadata.update(**self.extra_run_table_fields)
        run_metadata["run_id"] = dbt_run_id
        run_tbl = Table([run_metadata])

        node_rows = []
        for result in manifest.results:
            node_row = {"dbt_run_id": dbt_run_id}
            node_row.update(
                {
                    key: value
                    for key, value in result.__dict__.items()
                    if key in ("execution_time", "message")
                }
            )
            node_row["status"] = str(result.status)
            node_info = {
                key: value
                for key, value in result.node.__dict__.items()
                if key in ("database", "schema", "name", "path")
            }
            node_info["node"] = result.node.unique_id
            node_row.update(node_info)

            adapter_response_data = {
                key: value
                for key, value in result.adapter_response.items()
                if key in ("bytes_processed", "bytes_billed", "job_id", "slot_ms")
            }
            node_row.update(adapter_response_data)

            node_rows.append(node_row)

        nodes_tbl = Table(node_rows)
        return run_tbl, nodes_tbl

    def format_result(self) -> tuple[Table, Table]:
        """Returns a table for the dbt runs and a table for the node runs."""
        run_rows = []
        node_rows = []
        for command in self.commands:
            run_tbl, nodes_tbl = self.format_command_result(command)
            run_rows.extend(run_tbl.to_dicts())
            node_rows.extend(nodes_tbl.to_dicts())

        all_runs_tbl = Table(run_rows)
        all_nodes_tbl = Table(node_rows)
        return all_runs_tbl, all_nodes_tbl

    def send(self, manifests: list[Manifest]) -> None:
        self.commands = manifests
        runs_tbl, nodes_tbl = self.format_result()

        if len(runs_tbl):
            self.db_connector.copy(
                runs_tbl,
                self.destination_table_runs,
                if_exists="append",
                **self.copy_kwargs,
            )
        if len(nodes_tbl):
            self.db_connector.copy(
                nodes_tbl,
                self.destination_table_nodes,
                if_exists="append",
                **self.copy_kwargs,
            )



import logging

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

# Set up logger
logger = logging.getLogger(__name__)

QB_URI = "https://rest.tsheets.com/api/v1/"


class QuickBooksTime:
    """
    Instantiate the QuickBooksTime class.

    `Args:`
        token: str
            A valid QuickBooksTime Auth Token. Not required if ``QB_AUTH_TOKEN`` env
            variable set.
            [Find instructions to create yours here](https://developer.intuit.com/app/developer/qbo/docs/develop/authentication-and-authorization/oauth-2.0) # noqa: E501

            [QuickBooksTime API Documentation](https://tsheetsteam.github.io/api_docs/#introduction)
    """

    def __init__(self, token=None):
        self.token = check_env.check("QB_AUTH_TOKEN", token)
        self.headers = {"Authorization": "Bearer " + self.token}
        self.client = APIConnector(QB_URI, headers=self.headers)

    # Helper functions

    def qb_get_request(self, end_point: str, querystring=None):
        """
        This function handles the pagination of the request
        """

        # If no querystring is provided, initialize it as an empty dictionary
        if querystring is None:
            querystring = {}

        output_list = []  # This list will hold the results

        # Handle page parameter
        page = querystring.get("page", 1)

        more = True  # This flag indicates if there are more pages to fetch
        while more:
            # After every 10 pages, log the progress
            if page % 10 == 0:
                logger.info(f"Retrieved {len(output_list)} records from {end_point} endpoint.")
                logger.info(f"Currently on page {page}.")

            # Add the current page to the querystring
            querystring = {**querystring, **{"page": page}}

            # Send the GET request
            response = self.client.get_request(end_point, params=querystring)

            # Extract the key of the results
            endpoint_key = list(response["results"].keys())[0]

            # Extract the records from the results
            temp_list = list(response["results"][endpoint_key].values())

            # If the response indicates there are more pages,
            # update the flag and increment the page number
            more = response.get("more", False)
            page += 1

            # Add the records from the current page to the output list
            output_list.extend(temp_list)

        # Log the total number of records retrieved
        logger.info(f"Retrieved {len(output_list)} records from {end_point} endpoint.")

        # Return the results as a Table
        return Table(output_list)

    def get_groups(
        self,
        ids=None,
        active=None,
        manager_ids=None,
        name=None,
        modified_before=None,
        modified_since=None,
        supplemental_data=None,
        limit=None,
        page=1,
    ):
        """
            This function allows you to call the /groups endpoint of the QuickBooksTime Time API.
            All Args are optional.

            `Args:`
                ids: Int
                    Comma separated list of one or more group ids you'd like to filter on.

                active: String
                    'yes', 'no', or 'both'. Default is 'yes'.

                manager_ids: Int
                    Comma separated list of one or more manager ids you'd like to filter on.

                name: String
                    Comma separated list of one or more group names you'd like to filter on.

                modified_before: String
                    Only groups modified before this date/time will be returned,
                    in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm)

                modified_since: String
                    Only groups modified since this date/time will be returned,
                    in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm)

                supplemental_data: String
                    'yes' or 'no'. Default is 'yes'.
                    Indicates whether supplemental data should be returned.

                limit: Int
                    Represents how many results you'd like to retrieve per request (page).
                    Default is 200. Max is 200.

                page: Int
                    Represents the page of results you'd like to retrieve.

        `Returns:`
            Parsons Table
        """

        querystring = {
            "ids": ids,
            "active": active,
            "manager_ids": manager_ids,
            "name": name,
            "modified_before": modified_before,
            "modified_since": modified_since,
            "supplemental_data": supplemental_data,
            "limit": limit,
            "page": page,
        }

        logger.info("Retrieving groups.")
        tbl = self.qb_get_request(end_point="groups", querystring=querystring)

        logger.info(f"Found {tbl.num_rows} groups.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_jobcodes(
        self,
        ids=None,
        parent_ids=None,
        name=None,
        type=None,
        active=None,
        customfields=None,
        modified_before=None,
        modified_since=None,
        supplemental_data=None,
        limit=None,
        page=1,
    ):
        """
        This function allows you to call the /jobcodes endpoint of the QuickBooksTime Time API.

        `Args:`
            ids: Int
                Comma separated list of one or more jobcode ids you'd like to filter on.
                Only jobcodes with an id set to one of these values will be returned.
                If omitted, all jobcodes matching other specified filters are returned.

            parent_ids: Int
                Default is -1 (meaning all jobcodes will be returned regardless of parent_id).
                Comma separated list of one or more jobcode parent_ids you'd like to filter on.
                Only jobcodes with a parent_id set to one of these values will be returned.
                Additionally you can use 0 to get only the top-level jobcodes.

                Then get the id of any results with has_children=yes
                and feed that in as the value of parent_ids
                for your next request to get the 2nd level of jobcodes,
                and so on, to traverse an entire tree of jobcodes.
                Use -1 to return all jobcodes regardless of parent_id.
                This is especially useful when combined with the modified_since filter.
                When parent_ids is -1, you'll have the jobcode records needed to trace each result
                back to it's top level parent in the supplemental_data section of the response.

            name: String
                * will be interpreted as a wild card.
                Starts matching from the beginning of the string.

            type: String
                Indicates jobcode type. One of 'regular', 'pto', 'paid_break', 'unpaid_break',
                or 'all'. Default is 'regular'.

            active: String
                'yes', 'no', or 'both'. Default is 'yes'. If a jobcode is active,
                it is available for selection during time entry.

            customfields: Boolean
                true or false. If true, custom fields for this jobcode will be returned.
                If false, the customfields object will be omitted.

            modified_before: String
                Only jobcodes modified before this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            modified_since: String
                Only jobcodes modified since this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            supplemental_data: String
                'yes' or 'no'. Default is 'yes'.
                Indicates whether supplemental data should be returned.

            limit: Int
                Represents how many results you'd like to retrieve per request (page).
                Default is 200. Max is 200.

            page: Int
                Represents the page of results you'd like to retrieve. Default is 1.

        `Returns:`
            Parsons Table
        """

        querystring = {
            "ids": ids,
            "parent_ids": parent_ids,
            "name": name,
            "customfields": customfields,
            "active": active,
            "type": type,
            "modified_before": modified_before,
            "modified_since": modified_since,
            "supplemental_data": supplemental_data,
            "limit": limit,
            "page": page,
        }

        logger.info("Retrieving jobcodes.")
        tbl = self.qb_get_request(end_point="jobcodes", querystring=querystring)

        logger.info(f"Found {tbl.num_rows} jobs.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_timesheets(
        self,
        ids=None,
        jobcode_ids=None,
        payroll_ids=None,
        user_ids=None,
        group_ids=None,
        end_date=None,
        on_the_clock=None,
        jobcode_type=None,
        modified_before=None,
        modified_since=None,
        supplemental_data=None,
        limit=None,
        page=1,
        start_date="1900-01-01",
    ):
        """
        This function allows you to call the /timesheets endpoint of the QuickBooksTime Time API.

        Params:
            ids: Int
                required (unless modified_before, modified_since, or start_date are set)
                Comma separated list of one or more timesheet ids you'd like to filter on.
                Only timesheets with an id set to one of these values will be returned.
                If omitted, all timesheets matching other specified filters are returned.

            start_date: String
                required (unless modified_before, modified_since, or ids is set)
                YYYY-MM-DD formatted date.
                Any timesheets with a date falling on or after this date will be returned.

            end_date: String
                YYYY-MM-DD formatted date.
                Any timesheets with a date falling on or before this date will be returned.

            jobcode_ids: Int
                A comma-separated string of jobcode ids.
                Only time recorded against these jobcodes or one of their children will be returned.

            payroll_ids: Int
                A comma-separated string of payroll ids.
                Only time recorded against users with these payroll ids will be returned.

            user_ids: Int
                A comma-separated list of user ids.
                Only timesheets linked to these users will be returned.

            group_ids: Int
                A comma-separated list of group ids.
                Only timesheets linked to users from these groups will be returned.

            on_the_clock: String
                'yes', 'no', or 'both'. Default is 'no'.
                If a timesheet is on_the_clock, it means the user is currently working
                (has not clocked out yet).

            jobcode_type: String
                'regular', 'pto', 'paid_break', 'unpaid_break', or 'all'. Default is 'all'.
                Only timesheets linked to a jobcode of the given type are returned.

            modified_before: String
                required (unless modified_since, ids, or start_date are set)
                Only timesheets modified before this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            modified_since: String
                required (unless modified_before, ids, or start_date are set)
                Only timesheets modified since this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            supplemental_data: String
                'yes' or 'no'. Default is 'yes'.
                Indicates whether supplemental data should be returned.

            limit: Int
                Represents how many results you'd like to retrieve per request (page).
                Default is 200. Max is 200.

            page: Int
                Represents the page of results you'd like to retrieve. Default is 1.

        `Returns:`
            Parsons Table
        """

        querystring = {
            "ids": ids,
            "jobcode_ids": jobcode_ids,
            "payroll_ids": payroll_ids,
            "user_ids": user_ids,
            "group_ids": group_ids,
            "end_date": end_date,
            "on_the_clock": on_the_clock,
            "jobcode_type": jobcode_type,
            "modified_before": modified_before,
            "modified_since": modified_since,
            "supplemental_data": supplemental_data,
            "limit": limit,
            "start_date": start_date,
            "page": page,
        }

        logger.info("Retrieving timesheets.")
        tbl = self.qb_get_request(end_point="timesheets", querystring=querystring)

        logger.info(f"Found {tbl.num_rows} timesheets.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_users(
        self,
        ids=None,
        not_ids=None,
        employee_numbers=None,
        usernames=None,
        group_ids=None,
        not_group_ids=None,
        payroll_ids=None,
        active=None,
        first_name=None,
        last_name=None,
        modified_before=None,
        modified_since=None,
        supplemental_data=None,
        limit=None,
        page=1,
    ):
        """
        This function allows you to call the /users endpoint of the QuickBooksTime Time API.

        `Args:`
            ids: Int
                Comma separated list of one or more user ids you'd like to filter on.

            not_ids: Int
                Comma separated list of one or more user ids you'd like to filter on.
                Specifically, the user ids you'd like to exclude.

            employee_numbers: Int
                Comma separated list of one or more employee numbers you'd like to filter on.

            usernames: Str
                Comma separated list of one or more usernames you'd like to filter on.

            group_ids: Int
                Comma separated list of one or more group ids you'd like to filter on.

            not_group_ids: Int
                Comma separated list of one or more group ids you'd like to filter on.
                Specifically, the group ids you'd like to exclude.

            payroll_ids: String
                A comma-separated string of payroll ids.
                Only users with these payroll ids will be returned.

            active: String
                'yes', 'no', or 'both'. Default is 'yes'.

            first_name: String
                * will be interpreted as a wild card.
                Starts matching from the beginning of the string.

            last_name: String
                * will be interpreted as a wild card.
                Starts matching from the beginning of the string.

            modified_before: String
                Only users modified before this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            modified_since: String
                Only users modified since this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            supplemental_data: String
                'yes' or 'no'. Default is 'yes'.
                Indicates whether supplemental data should be returned.

            limit: Int
                Represents how many results you'd like to retrieve per request (page).
                Default is 200. Max is 200.

            page: Int
                Represents the page of results you'd like to retrieve. Default is 1.

        `Returns:`
            Parsons Table
            See Parsons Table for output options.
        """

        querystring = {
            "ids": ids,
            "not_ids": not_ids,
            "employee_numbers": employee_numbers,
            "usernames": usernames,
            "group_ids": group_ids,
            "not_group_ids": not_group_ids,
            "payroll_ids": payroll_ids,
            "active": active,
            "first_name": first_name,
            "last_name": last_name,
            "modified_before": modified_before,
            "modified_since": modified_since,
            "supplemental_data": supplemental_data,
            "limit": limit,
            "page": page,
        }

        logger.info("Retrieving users.")
        tbl = self.qb_get_request(end_point="users", querystring=querystring)

        logger.info(f"Found {tbl.num_rows} users.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_schedule_calendars_list(
        self,
        ids=None,
        modified_before=None,
        modified_since=None,
        supplemental_data=None,
        limit=None,
        page=1,
    ):
        """
        This function allows you to call the /schedule_calendars endpoint
        of the QuickBooksTime Time API.

        `Args:`
            ids: Int
                Comma separated list of one or more schedule calendar ids you'd like to filter on.
                Only schedule calendars with an id set to one of these values will be returned.

            modified_before: String
                Only schedule calendars modified before this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            modified_since: String
                required (unless ids, modified_before, or start are set)
                Only schedule calendars modified since this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            supplemental_data: String
                'yes' or 'no'. Default is 'yes'.
                Indicates whether supplemental data should be returned.

            limit: Int
                Represents how many results you'd like to retrieve per request (page).
                Default is 200. Max is 200.

            page: Int
                Represents the page of results you'd like to retrieve. Default is 1.

        `Returns:`
            List of integers of schedules calendar ids.
            Needed for calling the /schedule_events endpoint
        """

        endpoint = "schedule_calendars"

        querystring = {
            "ids": ids,
            "modified_before": modified_before,
            "modified_since": modified_since,
            "supplemental_data": supplemental_data,
            "limit": limit,
            "page": page,
        }

        logger.info("Retrieving schedule calendars.")
        tbl = self.qb_get_request(end_point=endpoint, querystring=querystring)

        schedule_calendar_ids_list = [
            int(row["id"]) for row in tbl
        ]  # Creates list of integers of the schedule_calendar_ids

        logger.info(f"Found {tbl.num_rows} schedule calendars.")

        return schedule_calendar_ids_list

    def get_schedule_events(
        self,
        ids=None,
        users_ids=None,
        schedule_calendar_ids=None,
        jobcode_ids=None,
        start="1970-01-01T00:00:00+00:00",
        end=None,
        active_users=None,
        active=None,
        draft=None,
        team_events=None,
        modified_before=None,
        modified_since=None,
        supplemental_data=None,
        limit=None,
        page=1,
    ):
        """
        This function allows you to call the /schedule_events endpoint
        of the QuickBooksTime Time API.

        `Args:`
            ids: Int
                required (unless modified_before, modified_since, or start are set)
                Comma separated list of one or more schedule event ids you'd like to filter on.
                Only schedule events with an id set to one of these values will be returned.

            users_ids: Int
                Comma-separated list of one or more user ids to retrieve schedule events for.

            schedule_calendar_ids: Int
                Required.
                Comma separated list of one or more schedule calendar ids you'd like to filter on.
                Only schedule events with a schedule calendar id
                set to one of these values will be returned.

            jobcode_ids: Int
                A comma-separated string of jobcode ids.
                Only schedule events with these jobcodes will be returned.

            start: String
                required (unless ids, modified_before, or modified_since are set)
                Only schedule events starting on or after this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            end: String
                Only schedule events ending on or before this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            active: String
                'yes', 'no' or 'both'. Default is 'both'.
                Only schedule events whose active state match the requested filter will be returned.

            active_users: Int
                '0', '-1' or '1' . Default is '1'.
                Only schedule events whose users are active will be returned by default.
                0 will return events for inactive users.
                -1 will return events for active and inactive users.

            draft: String
                'yes', 'no' or 'both'. Default is 'no'.
                Only schedule events whose draft state match the requested filter will be returned.

            team_events: String
                'base' or 'instance'. Default is 'instance'.
                If 'instance' is specified,
                events that are assigned to multiple users will be returned
                as individual single events for each assigned user. If 'base' is specified,
                events that are assigned to multiple users will be returned as one combined event
                for all assignees.

            modified_before: String
                required (unless ids, modified_since, or start are set)
                Only schedule events modified before this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            modified_since: String
                required (unless ids, modified_before, or start are set)
                Only schedule events modified since this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).

            supplemental_data: String
                'yes' or 'no'. Default is 'yes'.
                Indicates whether supplemental data should be returned.

            limit: Int
                Represents how many results you'd like to retrieve per request (page).
                Default is 200. Max is 200.

            page: Int
                Represents the page of results you'd like to retrieve. Default is 1.

        `Returns:`
            Parsons Table
            See Parsons Table for output options.
        """

        if schedule_calendar_ids is None:
            schedule_calendar_ids = self.get_schedule_calendars_list()

        endpoint = "schedule_events"

        querystring = {
            "ids": ids,
            "users_ids": users_ids,
            "schedule_calendar_ids": schedule_calendar_ids,
            "jobcode_ids": jobcode_ids,
            "start": start,
            "end": end,
            "active": active,
            "active_users": active_users,
            "draft": draft,
            "team_events": team_events,
            "modified_before": modified_before,
            "modified_since": modified_since,
            "supplemental_data": supplemental_data,
            "limit": limit,
            "page": page,
        }

        logger.info("Retrieving schedule events.")
        tbl = self.qb_get_request(end_point=endpoint, querystring=querystring)

        logger.info(f"Found {tbl.num_rows} schedules.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_geolocations(
        self,
        ids=None,
        modified_before=None,
        modified_since=None,
        user_ids=None,
        group_ids=None,
        supplemental_data=None,
        limit=None,
        page=1,
    ):
        """
        This function allows you to call the /geolocations endpoint of the QuickBooksTime Time API.

        `Args:`
            ids: Int
                Comma separated list of one or more geolocation ids you'd like to filter on.
                Only geolocations with an id set to one of these values will be returned.
                Required (unless modified_before, modified_since is set)
            modified_before: String
                Only geolocations modified before this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).
                Required (unless ids or modified_since is set)
            modified_since: String
                Only geolocations modified since this date/time will be returned,
                in ISO 8601 format (YYYY-MM-DDThh:mm:ss±hh:mm).
                Required (unless ids or modified_before is set)
            user_ids: Int
                Comma separated list of one or more user ids you'd like to filter on.
                Only geolocations with a user id set to one of these values will be returned.
            group_ids: Int
                Comma separated list of one or more group ids you'd like to filter on.
                Only geolocations with a group id set to one of these values will be returned.
            suplemental_data: String
                'yes' or 'no'. Default is 'yes'.
                Indicates whether supplemental data should be returned.
            limit: Int
                Represents how many results you'd like to retrieve per request (page).
                Default is 200. Max is 200.
            page: Int
                Represents the page of results you'd like to retrieve. Default is 1.

        `Returns:`
            Parsons Table
            See Parsons Table for output options.
        """

        endpoint = "geolocations"

        querystring = {
            "ids": ids,
            "modified_before": modified_before,
            "modified_since": modified_since,
            "user_ids": user_ids,
            "group_ids": group_ids,
            "supplemental_data": supplemental_data,
            "limit": limit,
            "page": page,
        }

        logger.info("Retrieving geolocations.")
        tbl = self.qb_get_request(end_point=endpoint, querystring=querystring)

        logger.info(f"Found {tbl.num_rows} geolocations.")

        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()



import json
import logging
import os
import unittest.mock as mock
from typing import Union
from unittest import TestCase
from unittest.mock import Mock

from google.cloud import bigquery, exceptions
from testfixtures import log_capture

from parsons import GoogleBigQuery, Table
from parsons.google.google_cloud_storage import GoogleCloudStorage
from test.test_google.test_utilities import FakeCredentialTest


class BigQuery(GoogleBigQuery):
    @mock.patch("parsons.google.google_bigquery.load_google_application_credentials")
    def __init__(self, load_creds_mock, app_creds=None, **kwargs):
        super().__init__(app_creds=app_creds, **kwargs)


class FakeClient:
    """A Fake Storage Client used for monkey-patching."""

    @mock.patch("parsons.google.google_bigquery.load_google_application_credentials")
    @mock.patch("parsons.google.google_cloud_storage.load_google_application_credentials")
    def __init__(self, load_creds_mock, load_creds_mock_2, project=None, credentials=None):
        self.project = project


class FakeGoogleCloudStorage(GoogleCloudStorage):
    """A Fake GoogleCloudStorage object used to test setting up credentials."""

    @mock.patch("google.cloud.storage.Client", FakeClient)
    @mock.patch("parsons.google.google_cloud_storage.load_google_application_credentials")
    def __init__(self, load_creds_mock):
        super().__init__(None, None)

    def upload_table(self, table, bucket_name, blob_name, data_type="csv", default_acl=None):
        pass

    def delete_blob(self, bucket_name, blob_name):
        pass


class TestGoogleBigQuery(FakeCredentialTest):
    def setUp(self):
        super().setUp()
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.cred_path
        self.tmp_gcs_bucket = "tmp"

    def tearDown(self) -> None:
        super().tearDown()
        del os.environ["GOOGLE_APPLICATION_CREDENTIALS"]

    def test_query(self):
        query_string = "select * from table"

        # Pass the mock class into our GoogleBigQuery constructor
        bq = self._build_mock_client_for_querying([{"one": 1, "two": 2}])

        # Run a query against our parsons GoogleBigQuery class
        result = bq.query(query_string)

        # Check our return value
        self.assertEqual(result.num_rows, 1)
        self.assertEqual(result.columns, ["one", "two"])
        self.assertEqual(result[0], {"one": 1, "two": 2})

    def test_query__no_results(self):
        query_string = "select * from table limit 0"

        # Pass the mock class into our GoogleBigQuery constructor
        bq = self._build_mock_client_for_querying([])

        # Run a query against our parsons GoogleBigQuery class
        result = bq.query(query_string)

        # Check our return value
        # We can't use assertEqual(result, Table())
        # Because Table() == Table() fails for some reason
        assert isinstance(result, Table)
        assert not len(result)
        assert tuple(result.columns) == ()

    @mock.patch("parsons.utilities.files.create_temp_file")
    def test_query__no_return(self, create_temp_file_mock):
        query_string = "select * from table"

        # Pass the mock class into our GoogleBigQuery constructor
        bq = self._build_mock_client_for_querying([{"one": 1, "two": 2}])
        bq._fetch_query_results = mock.MagicMock()

        # Run a query against our parsons GoogleBigQuery class
        result = bq.query(query_string, return_values=False)

        # Check our return value
        self.assertEqual(result, None)

        # Check that query results were not fetched
        bq._fetch_query_results.assert_not_called()

    @mock.patch("parsons.utilities.files.create_temp_file")
    def test_query_with_transaction(self, create_temp_file_mock):
        queries = ["select * from table", "select foo from bar"]
        parameters = ["baz"]

        # Pass the mock class into our GoogleBigQuery constructor
        bq = self._build_mock_client_for_querying([{"one": 1, "two": 2}])
        bq.query = mock.MagicMock()

        # Run a query against our parsons GoogleBigQuery class
        result = bq.query_with_transaction(queries=queries, parameters=parameters)
        keyword_args = bq.query.call_args[1]

        # Check our return value
        self.assertEqual(result, None)

        # Check that queries and transaction keywords are included in sql
        self.assertTrue(
            all(text in keyword_args["sql"] for text in queries + ["BEGIN TRANSACTION", "COMMIT"])
        )
        self.assertEqual(keyword_args["parameters"], parameters)
        self.assertFalse(keyword_args["return_values"])

    def test_extract(self):
        gcs_bucket = "tmp"
        gcs_blob_name = "file/*"
        dataset = "dataset"
        table_name = "table"
        gs_tmp_destination = f"gs://{gcs_bucket}/{gcs_blob_name}"
        bq = self._build_mock_client_for_copying(table_exists=False)
        bq.extract(
            gcs_bucket=gcs_bucket,
            gcs_blob_name=gcs_blob_name,
            dataset=dataset,
            table_name=table_name,
        )

        self.assertEqual(bq.client.extract_table.call_count, 1)
        load_call_args = bq.client.extract_table.call_args
        self.assertEqual(load_call_args[1]["destination_uris"], gs_tmp_destination)

        job_config = load_call_args[1]["job_config"]
        self.assertEqual(job_config.destination_format, bigquery.DestinationFormat.CSV)

    def test_get_job(self):
        tmp_job_id = "1234567890"
        bq = self._build_mock_base_client()
        bq.client.get_job(job_id=tmp_job_id)

        self.assertEqual(bq.client.get_job.call_count, 1)
        load_call_args = bq.client.get_job.call_args
        self.assertEqual(load_call_args[1]["job_id"], tmp_job_id)

    def test_copy_gcs(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)

        # call the method being tested
        bq.copy_from_gcs(
            gcs_blob_uri=tmp_blob_uri,
            table_name="dataset.table",
        )

        # check that the method did the right things
        self.assertEqual(bq.client.load_table_from_uri.call_count, 1)
        load_call_args = bq.client.load_table_from_uri.call_args
        self.assertEqual(load_call_args[1]["source_uris"], tmp_blob_uri)

        job_config = load_call_args[1]["job_config"]
        self.assertEqual(job_config.write_disposition, bigquery.WriteDisposition.WRITE_EMPTY)

    def test_copy_gcs__if_exists_truncate(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)

        # call the method being tested
        bq.copy_from_gcs(
            gcs_blob_uri=tmp_blob_uri,
            table_name="dataset.table",
            if_exists="truncate",
        )

        # check that the method did the right things
        self.assertEqual(bq.client.load_table_from_uri.call_count, 1)
        load_call_args = bq.client.load_table_from_uri.call_args
        self.assertEqual(load_call_args[1]["source_uris"], tmp_blob_uri)

        job_config = load_call_args[1]["job_config"]
        self.assertEqual(job_config.write_disposition, bigquery.WriteDisposition.WRITE_TRUNCATE)

    def test_copy_gcs__if_exists_append(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)

        # call the method being tested
        bq.copy_from_gcs(
            gcs_blob_uri=tmp_blob_uri,
            table_name="dataset.table",
            if_exists="append",
        )

        # check that the method did the right things
        self.assertEqual(bq.client.load_table_from_uri.call_count, 1)
        load_call_args = bq.client.load_table_from_uri.call_args
        self.assertEqual(load_call_args[1]["source_uris"], tmp_blob_uri)

        job_config = load_call_args[1]["job_config"]
        self.assertEqual(job_config.write_disposition, bigquery.WriteDisposition.WRITE_APPEND)

    def test_copy_gcs__if_exists_fail(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)

        # call the method being tested
        bq.copy_from_gcs(
            gcs_blob_uri=tmp_blob_uri,
            table_name="dataset.table",
            if_exists="truncate",
        )
        bq.table_exists = mock.MagicMock()
        bq.table_exists.return_value = True

        # call the method being tested
        with self.assertRaises(Exception):  # noqa: B017
            bq.copy_from_gcs(
                self.default_table,
                "dataset.table",
                tmp_gcs_bucket=self.tmp_gcs_bucket,
                gcs_client=self._build_mock_cloud_storage_client(),
            )

    def test_copy_gcs__if_exists_drop(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)
        bq.table_exists = mock.MagicMock()
        bq.table_exists.return_value = True

        # call the method being tested
        bq.copy_from_gcs(
            gcs_blob_uri=tmp_blob_uri,
            table_name="dataset.table",
            if_exists="drop",
        )

        # check that we tried to delete the table
        self.assertEqual(bq.client.delete_table.call_count, 1)

    def test_copy_gcs__bad_if_exists(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)
        bq.table_exists = mock.MagicMock()
        bq.table_exists.return_value = True

        # call the method being tested
        with self.assertRaises(ValueError):
            bq.copy_from_gcs(
                gcs_blob_uri=tmp_blob_uri,
                table_name="dataset.table",
                if_exists="foobar",
            )

    @mock.patch("google.cloud.storage.Client")
    @mock.patch("parsons.google.google_cloud_storage.load_google_application_credentials")
    @mock.patch.object(GoogleCloudStorage, "split_uri", return_value=("tmp", "file.gzip"))
    @mock.patch.object(GoogleCloudStorage, "unzip_blob", return_value="gs://tmp/file.csv")
    def test_copy_large_compressed_file_from_gcs(
        self, unzip_mock: mock.MagicMock, split_mock: mock.MagicMock, *_
    ):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file.gzip"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)

        # call the method being tested
        bq.copy_large_compressed_file_from_gcs(
            gcs_blob_uri=tmp_blob_uri,
            table_name="dataset.table",
        )

        # check that the method did the right things
        split_mock.assert_has_calls(
            [
                mock.call(gcs_uri="gs://tmp/file.gzip"),
                mock.call(gcs_uri="gs://tmp/file.csv"),
            ]
        )
        unzip_mock.assert_called_once_with(
            bucket_name="tmp",
            blob_name="file.gzip",
            new_file_extension="csv",
            compression_type="gzip",
        )
        self.assertEqual(bq.client.load_table_from_uri.call_count, 1)
        load_call_args = bq.client.load_table_from_uri.call_args
        self.assertEqual(load_call_args[1]["source_uris"], "gs://tmp/file.csv")

        job_config = load_call_args[1]["job_config"]
        self.assertEqual(job_config.write_disposition, bigquery.WriteDisposition.WRITE_EMPTY)

    def test_copy_s3(self):
        # setup dependencies / inputs
        table_name = "table_name"
        bucket = "aws_bucket"
        key = "file.gzip"
        aws_access_key_id = "AAAAAA"
        aws_secret_access_key = "BBBBB"
        tmp_gcs_bucket = "tmp"

        # set up object under test
        bq = self._build_mock_client_for_copying(table_exists=False)
        gcs_client = self._build_mock_cloud_storage_client()
        bq.copy_from_gcs = mock.MagicMock()

        # call the method being tested
        bq.copy_s3(
            table_name=table_name,
            bucket=bucket,
            key=key,
            gcs_client=gcs_client,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            tmp_gcs_bucket=tmp_gcs_bucket,
        )

        # check that the method did the right things
        gcs_client.copy_s3_to_gcs.assert_called_once_with(
            aws_source_bucket=bucket,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            gcs_sink_bucket=tmp_gcs_bucket,
            aws_s3_key=key,
        )
        bq.copy_from_gcs.assert_called_once()
        gcs_client.delete_blob.assert_called_once()

    def test_copy(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        gcs_client = self._build_mock_cloud_storage_client(tmp_blob_uri)
        tbl = self.default_table
        bq = self._build_mock_client_for_copying(table_exists=False)
        bq._load_table_from_uri = mock.MagicMock()
        bq.get_table_ref = mock.Mock(wraps=bq.get_table_ref)
        table_name = "dataset.table"

        # call the method being tested
        bq.copy(
            tbl,
            table_name,
            tmp_gcs_bucket=self.tmp_gcs_bucket,
            gcs_client=gcs_client,
        )

        # check that the method did the right things
        self.assertEqual(gcs_client.upload_table.call_count, 1)
        upload_call_args = gcs_client.upload_table.call_args
        self.assertEqual(upload_call_args[0][0], tbl)
        self.assertEqual(upload_call_args[0][1], self.tmp_gcs_bucket)
        tmp_blob_name = upload_call_args[0][2]

        self.assertEqual(bq._load_table_from_uri.call_count, 1)
        load_call_args = bq._load_table_from_uri.call_args
        job_config = load_call_args[1]["job_config"]
        column_types = [schema_field.field_type for schema_field in job_config.schema]
        self.assertEqual(column_types, ["INTEGER", "STRING", "BOOLEAN"])
        self.assertEqual(load_call_args[1]["source_uris"], tmp_blob_uri)

        self.assertEqual(bq.get_table_ref.call_count, 2)
        get_table_ref_args = bq.get_table_ref.call_args
        self.assertEqual(get_table_ref_args[1]["table_name"], table_name)

        # make sure we cleaned up the temp file
        self.assertEqual(gcs_client.delete_blob.call_count, 1)
        delete_call_args = gcs_client.delete_blob.call_args
        self.assertEqual(delete_call_args[0][0], self.tmp_gcs_bucket)
        self.assertEqual(delete_call_args[0][1], tmp_blob_name)

    @mock.patch("parsons.google.google_cloud_storage.load_google_application_credentials")
    @mock.patch("parsons.google.google_bigquery.load_google_application_credentials")
    def test_copy__credentials_are_correctly_set__from_filepath(
        self, load_creds_mock, load_creds_mock_2
    ):
        tbl = self.default_table
        bq = self._build_mock_client_for_copying(table_exists=False, app_creds=self.cred_path)

        # Pass in our fake GCS Client.
        bq.copy(
            tbl,
            "dataset.table",
            tmp_gcs_bucket=self.tmp_gcs_bucket,
            gcs_client=FakeGoogleCloudStorage(),
        )

        actual = os.environ[bq.env_credential_path]

        with open(actual, "r") as factual:
            with open(self.cred_path, "r") as fexpected:
                actual_str = factual.read()
                self.assertEqual(actual_str, fexpected.read())
                self.assertEqual(self.cred_contents, json.loads(actual_str))

    @mock.patch("parsons.google.google_cloud_storage.load_google_application_credentials")
    @mock.patch("parsons.google.google_bigquery.load_google_application_credentials")
    def test_copy__credentials_are_correctly_set__from_env(
        self, load_creds_mock, load_creds_mock_2
    ):
        tbl = self.default_table
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.cred_path

        bq = self._build_mock_client_for_copying(
            table_exists=False,
        )

        # Pass in our fake GCS Client.
        bq.copy(
            tbl,
            "dataset.table",
            tmp_gcs_bucket=self.tmp_gcs_bucket,
            gcs_client=FakeGoogleCloudStorage(),
        )

        actual = os.environ[bq.env_credential_path]

        with open(actual, "r") as factual:
            with open(self.cred_path, "r") as fexpected:
                actual_str = factual.read()
                self.assertEqual(actual_str, fexpected.read())
                self.assertEqual(self.cred_contents, json.loads(actual_str))

    @mock.patch("parsons.google.google_cloud_storage.load_google_application_credentials")
    @mock.patch("parsons.google.google_bigquery.load_google_application_credentials")
    def test_copy__credentials_are_correctly_set__from_dict(
        self, load_creds_mock, load_creds_mock_2
    ):
        tbl = self.default_table
        with open(self.cred_path) as file:
            cred_dict = json.loads(file.read())
        bq = self._build_mock_client_for_copying(table_exists=False, app_creds=cred_dict)

        # Pass in our fake GCS Client.
        bq.copy(
            tbl,
            "dataset.table",
            tmp_gcs_bucket=self.tmp_gcs_bucket,
            gcs_client=FakeGoogleCloudStorage(),
        )

        actual = os.environ[bq.env_credential_path]

        with open(actual, "r") as factual:
            with open(self.cred_path, "r") as fexpected:
                actual_str = factual.read()
                self.assertEqual(actual_str, fexpected.read())
                self.assertEqual(self.cred_contents, json.loads(actual_str))

    def test_copy__if_exists_passed_through(self):
        # setup dependencies / inputs
        tmp_blob_uri = "gs://tmp/file"

        # set up object under test
        gcs_client = self._build_mock_cloud_storage_client(tmp_blob_uri)
        tbl = self.default_table
        bq = self._build_mock_client_for_copying(table_exists=False)
        bq._load_table_from_uri = mock.MagicMock()
        bq._process_job_config = mock.Mock(wraps=bq._process_job_config)
        table_name = "dataset.table"
        if_exists = "drop"

        # call the method being tested
        bq.copy(
            tbl,
            table_name,
            tmp_gcs_bucket=self.tmp_gcs_bucket,
            gcs_client=gcs_client,
            if_exists=if_exists,
        )

        self.assertEqual(bq._load_table_from_uri.call_count, 1)
        process_job_config_args = bq._process_job_config.call_args
        self.assertEqual(process_job_config_args[1]["if_exists"], if_exists)

    @mock.patch.object(BigQuery, "table_exists", return_value=False)
    @mock.patch.object(BigQuery, "query", return_value=None)
    def test_duplicate_table(self, query_mock, table_exists_mock):
        source_table = "vendor_table"
        destination_table = "raw_table"
        expected_query = f"""
            CREATE TABLE
            {destination_table}
            CLONE {source_table}
        """
        bq = self._build_mock_client_for_querying(results=None)

        bq.duplicate_table(
            source_table=source_table,
            destination_table=destination_table,
        )

        query_mock.assert_called_once()
        actual_query = query_mock.call_args[1]["sql"]
        self.assertEqual(actual_query, expected_query)

    @mock.patch.object(BigQuery, "table_exists", return_value=False)
    @mock.patch.object(BigQuery, "delete_table", return_value=None)
    @mock.patch.object(BigQuery, "query", return_value=None)
    def test_duplicate_table_with_drop(
        self, query_mock: mock.MagicMock, delete_mock: mock.MagicMock, table_exists_mock
    ):
        source_table = "vendor_table"
        destination_table = "raw_table"
        bq = self._build_mock_client_for_querying(results=None)

        bq.duplicate_table(
            source_table=source_table,
            destination_table=destination_table,
            drop_source_table=True,
        )

        delete_mock.assert_called_once_with(table_name=source_table)

    @mock.patch.object(BigQuery, "table_exists", return_value=True)
    @mock.patch.object(BigQuery, "query_with_transaction", return_value=None)
    @mock.patch.object(BigQuery, "copy", return_value=None)
    def test_upsert(self, copy_mock, query_mock, *_):
        upsert_tbl = Table([["id", "name"], [1, "Jane"]])
        target_table = "my_dataset.my_target_table"
        primary_key = "id"
        bq = self._build_mock_client_for_querying(results=[])

        bq.upsert(
            table_obj=upsert_tbl,
            target_table=target_table,
            primary_key=primary_key,
            distinct_check=False,
        )

        # stages the table -> calls copy
        copy_mock.assert_called_once()
        self.assertEqual(copy_mock.call_args[1]["tbl"], upsert_tbl)
        self.assertEqual(copy_mock.call_args[1]["template_table"], target_table)

        # runs a delete insert within a transaction
        query_mock.assert_called_once()
        actual_queries = query_mock.call_args[1]["queries"]
        self.assertIn("DELETE", actual_queries[0])
        self.assertIn("INSERT", actual_queries[1])

    @mock.patch.object(BigQuery, "query")
    def test_get_row_count(self, query_mock):
        # Arrange
        schema = "foo"
        table_name = "bar"
        expected_num_rows = 2

        query_mock.return_value = Table([{"row_count": expected_num_rows}])
        expected_query = f"SELECT COUNT(*) AS row_count FROM `{schema}.{table_name}`"
        bq = self._build_mock_client_for_querying(results=Table([{"row_count": 2}]))

        # Act
        row_count = bq.get_row_count(schema=schema, table_name=table_name)

        # Assert
        query_mock.assert_called_once()
        actual_query = query_mock.call_args[1]["sql"]
        self.assertEqual(row_count, expected_num_rows)
        self.assertEqual(actual_query, expected_query)

    def _build_mock_client_for_querying(self, results):
        # Create a mock that will play the role of the cursor
        cursor = mock.MagicMock()
        cursor.execute.return_value = None
        cursor.fetchmany.side_effect = [results, []]
        if results:
            cursor.description = [(key, None) for key in results[0]]

        # Create a mock that will play the role of the connection
        connection = mock.MagicMock()
        connection.cursor.return_value = cursor

        # Create a mock that will play the role of the Google BigQuery dbapi module
        dbapi = mock.MagicMock()
        dbapi.connect.return_value = connection

        # Create a mock that will play the role of our GoogleBigQuery client
        client = mock.MagicMock()

        bq = BigQuery()
        bq._client = client
        bq._dbapi = dbapi
        return bq

    def _build_mock_client_for_copying(
        self, table_exists=True, app_creds: Union[str, dict, None] = None
    ):
        bq_client = mock.MagicMock()
        if not table_exists:
            bq_client.get_table.side_effect = exceptions.NotFound("not found")
        bq = BigQuery(app_creds=app_creds)
        bq._client = bq_client
        return bq

    def _build_mock_base_client(self, app_creds: Union[str, dict, None] = None):
        bq_client = mock.MagicMock()
        bq = BigQuery(app_creds=app_creds)
        bq._client = bq_client
        return bq

    def _build_mock_cloud_storage_client(self, tmp_blob_uri=""):
        gcs_client = mock.MagicMock()
        gcs_client.upload_table.return_value = tmp_blob_uri
        return gcs_client

    @property
    def default_table(self):
        return Table(
            [
                {"num": 1, "ltr": "a", "boolcol": None},
                {"num": 2, "ltr": "b", "boolcol": True},
            ]
        )


class TestGoogleBigQueryCopyBetweenProjects(TestCase):
    def setUp(self):
        # mock the GoogleBigQuery class
        self.bq = Mock(spec=GoogleBigQuery)

        # define inputs to copy method
        self.source_project = ("project1",)
        self.source_dataset = ("dataset1",)
        self.source_table = ("table1",)
        self.destination_project = ("project2",)
        self.destination_dataset = ("dataset2",)
        self.destination_table = ("table2",)
        self.if_dataset_not_exists = ("fail",)
        self.if_table_exists = "fail"

    def tearDown(self):
        pass

    def test_copy_called_once_with(self):
        self.bq.copy_between_projects(
            source_project=self.source_project,
            source_dataset=self.destination_dataset,
            source_table=self.source_table,
            destination_project=self.destination_project,
            destination_dataset=self.destination_dataset,
            destination_table=self.destination_table,
            if_dataset_not_exists=self.if_dataset_not_exists,
            if_table_exists=self.if_table_exists,
        )
        self.bq.copy_between_projects.assert_called_once_with(
            source_project=self.source_project,
            source_dataset=self.destination_dataset,
            source_table=self.source_table,
            destination_project=self.destination_project,
            destination_dataset=self.destination_dataset,
            destination_table=self.destination_table,
            if_dataset_not_exists=self.if_dataset_not_exists,
            if_table_exists=self.if_table_exists,
        )

    @log_capture()
    def test_logger_fail_on_dataset_does_not_exist(self, capture):
        # create and set up logger
        logger = logging.getLogger()
        logger.error(
            "Dataset {0} does not exist and if_dataset_not_exists set to {1}".format(
                self.destination_dataset, self.if_dataset_not_exists
            )
        )

        # call the method to generate log message
        self.bq.copy_between_projects(
            source_project=self.source_project,
            source_dataset=self.destination_dataset,
            source_table=self.source_table,
            destination_project=self.destination_project,
            destination_dataset=self.destination_dataset,
            destination_table=self.destination_table,
            if_dataset_not_exists=self.if_dataset_not_exists,
            if_table_exists=self.if_table_exists,
        )

        # check that the log message was generated correctly
        capture.check(
            (
                "root",
                "ERROR",
                "Dataset {0} does not exist and if_dataset_not_exists set to {1}".format(
                    self.destination_dataset, self.if_dataset_not_exists
                ),
            )
        )

    @log_capture()
    def test_logger_fail_on_table_exists(self, capture):
        # create and set up logger
        logger = logging.getLogger()

        ## now test with table copy error
        logger.error(
            "BigQuery copy failed, Table {0} exists and if_table_exists set to {1}".format(
                self.destination_table, self.if_table_exists
            )
        )

        # call the method to generate log message
        self.bq.copy_between_projects(
            source_project=self.source_project,
            source_dataset=self.destination_dataset,
            source_table=self.source_table,
            destination_project=self.destination_project,
            destination_dataset=self.destination_dataset,
            destination_table=self.destination_table,
            if_dataset_not_exists=self.if_dataset_not_exists,
            if_table_exists=self.if_table_exists,
        )

        # check that the log message was generated correctly
        capture.check(
            (
                "root",
                "ERROR",
                "BigQuery copy failed, Table {0} exists and if_table_exists set to {1}".format(
                    self.destination_table, self.if_table_exists
                ),
            )
        )



import os
import unittest

import requests_mock

from parsons import VAN

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestNGPVAN(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    def tearDown(self):
        pass

    @requests_mock.Mocker()
    def test_get_apikeyprofiles(self, m):
        json = {
            "items": [
                {
                    "databaseName": "SmartVAN Massachusetts",
                    "hasMyVoters": True,
                    "hasMyCampaign": True,
                    "committeeName": "People for Good",
                    "apiKeyTypeName": "Custom Integration",
                    "keyReference": "1234",
                    "userFirstName": "peopleforgood",
                    "userLastName": "api",
                    "username": "peopleforgood.api",
                    "userId": 4321,
                }
            ],
            "nextPageLink": None,
            "count": 1,
        }

        m.get(self.van.connection.uri + "apiKeyProfiles", json=json)

        # # Call the method that makes the API request
        response = self.van.get_apikeyprofiles()

        # Assert that the response is a dictionary (JSON object)
        self.assertIsInstance(response, dict)

        # Assert that the response matches the expected JSON
        # I have to access a part of the json because the response is a list of dictionaries
        # and the VAN Connector handles the pagination and unpacks the list of dictionaries
        self.assertEqual(response, json["items"][0])



"""Utility for running and logging output from dbt commands

Enable this utility by installing parsons with a dbt extra:
`pip install parsons[dbt-redshift]`
or `pip install parsons[dbt-postgres]`
or `pip install parsons[dbt-snowflake]`
or `pip install parsons[dbt-bigquery]`

To run dbt commands, you will need to have a dbt project directory
somewhere on the local filesystem.

The dbt command will inherit environment variables from the python
process shell, so if your dbt profiles.yml file uses environment
variables, ensure those are set in python or the parent shell before
running this dbt utility.

Logging is handled separately from the dbt run itself. The
dbtRunner.run method returns a dbtCommandResult object which can be
passed to a child class of dbtLogger for logging to stdout, slack,
etc.

Parsons provides a few example dbtLogger child classes, but for
best results, design your own!

Example usage:
```
from parsons.utilities.dbt import (
    run_dbt_commands,
    dbtLoggerSlack,
    dbtLoggerPython
)

run_dbt_commands(
    commands=['run', 'test'],
    dbt_project_directory='/home/ubuntu/code/dbt_project/',
    loggers=[
        dbtLoggerPython,
        dbtLoggerSlack(slack_webhook=os.environ['SLACK_WEBHOOK'])
    ]
)

```

"""

from parsons.utilities.dbt.dbt import run_dbt_commands
from parsons.utilities.dbt.logging import (
    dbtLoggerMarkdown,
    dbtLoggerPython,
    dbtLoggerSlack,
    dbtLoggerStdout,
)

__all__ = [
    "run_dbt_commands",
    "dbtLoggerMarkdown",
    "dbtLoggerSlack",
    "dbtLoggerStdout",
    "dbtLoggerPython",
]



import json
import os
import typing as t

import google
from google.oauth2 import service_account

from parsons.utilities import check_env, files


def setup_google_application_credentials(
    app_creds: t.Union[t.Dict, str, None],
    env_var_name: str = "GOOGLE_APPLICATION_CREDENTIALS",
    target_env_var_name: t.Optional[str] = None,
) -> None:
    # Detect if app_creds is a dict, path string or json string, and if it is a
    # json string, then convert it to a temporary file. Then set the
    # environmental variable.
    credentials = check_env.check(env_var_name, app_creds)
    try:
        if type(credentials) is dict:
            credentials = json.dumps(credentials)
        if json.loads(credentials):
            creds_path = files.string_to_temp_file(credentials, suffix=".json")
    except ValueError:
        creds_path = credentials

    if not target_env_var_name:
        target_env_var_name = env_var_name

    os.environ[target_env_var_name] = creds_path


def hexavigesimal(n: int) -> str:
    """
    Converts an integer value to the type of strings you see on spreadsheets
    (A, B,...,Z, AA, AB, ...).

    Code based on
    https://stackoverflow.com/questions/16190452/converting-from-number-to-hexavigesimal-letters

    `Args:`
        n: int
            A positive valued integer.

    `Returns:`
        str
            The hexavigeseimal representation of n
    """
    if n < 1:
        raise ValueError(f"This function only works for positive integers. Provided value {n}.")

    chars = ""
    while n != 0:
        chars = chr((n - 1) % 26 + 65) + chars  # 65 makes us start at A
        n = (n - 1) // 26
    return chars


def load_google_application_credentials(
    env_var_name: str = "GOOGLE_APPLICATION_CREDENTIALS",
    scopes: t.Optional[t.List[str]] = None,
    subject: t.Optional[str] = None,
) -> google.auth.credentials.Credentials:
    service_account_filepath = os.environ[env_var_name]

    credentials = service_account.Credentials.from_service_account_file(service_account_filepath)

    if scopes:
        credentials = credentials.with_scopes(scopes)

    if subject:
        credentials = credentials.with_subject(subject)

    return credentials



import unittest

import requests_mock

from parsons import Shopify, Table
from test.utils import assert_matching_tables

SUBDOMAIN = "myorg"
PASSWORD = "abc123"
API_KEY = "abc123"
API_VERSION = "2020-10"


class TestShopify(unittest.TestCase):
    mock_count_all = {"count": 2}
    mock_count_date = mock_count_since = {"count": 1}
    mock_graphql = {"data": {"orders": {"edges": [{"node": {"id": 1}}]}}}
    mock_orders_all = {
        "orders": [
            {
                "created_at": "2020-10-19T12:00:00-04:00",
                "financial_status": "paid",
                "id": 1,
            },
            {
                "created_at": "2020-10-20T12:00:00-04:00",
                "financial_status": "refunded",
                "id": 2,
            },
        ]
    }
    mock_orders_completed = {
        "orders": [
            {
                "created_at": "2020-10-19T12:00:00-04:00",
                "financial_status": "paid",
                "id": 1,
            }
        ]
    }
    mock_orders_date = mock_orders_since = {
        "orders": [
            {
                "created_at": "2020-10-20T12:00:00-04:00",
                "financial_status": "refunded",
                "id": 2,
            }
        ]
    }
    mock_result_all = Table(
        [
            ("created_at", "financial_status", "id"),
            ("2020-10-19T12:00:00-04:00", "paid", 1),
            ("2020-10-20T12:00:00-04:00", "refunded", 2),
        ]
    )
    mock_result_completed = Table(
        [
            ("created_at", "financial_status", "id"),
            ("2020-10-19T12:00:00-04:00", "paid", 1),
        ]
    )
    mock_result_date = mock_result_since = Table(
        [
            ("created_at", "financial_status", "id"),
            ("2020-10-20T12:00:00-04:00", "refunded", 2),
        ]
    )

    def setUp(self):
        self.shopify = Shopify(SUBDOMAIN, PASSWORD, API_KEY, API_VERSION)

    @requests_mock.Mocker()
    def test_get_count(self, m):
        m.get(
            self.shopify.get_query_url(None, None, "orders", True),
            json=self.mock_count_all,
        )
        m.get(
            self.shopify.get_query_url("2020-10-20", None, "orders", True),
            json=self.mock_count_date,
        )
        m.get(
            self.shopify.get_query_url(None, 2, "orders", True),
            json=self.mock_count_since,
        )
        self.assertEqual(self.shopify.get_count(None, None, "orders"), 2)
        self.assertEqual(self.shopify.get_count("2020-10-20", None, "orders"), 1)
        self.assertEqual(self.shopify.get_count(None, 2, "orders"), 1)

    @requests_mock.Mocker()
    def test_get_orders(self, m):
        m.get(
            self.shopify.get_query_url(None, None, "orders", False),
            json=self.mock_orders_all,
        )
        m.get(
            self.shopify.get_query_url("2020-10-20", None, "orders", False),
            json=self.mock_orders_date,
        )
        m.get(
            self.shopify.get_query_url(None, 2, "orders", False),
            json=self.mock_orders_since,
        )
        m.get(
            self.shopify.get_query_url(None, None, "orders", False) + "&financial_status=paid",
            json=self.mock_orders_completed,
        )
        assert_matching_tables(self.shopify.get_orders(None, None, False), self.mock_result_all)
        assert_matching_tables(
            self.shopify.get_orders("2020-10-20", None, False), self.mock_result_date
        )
        assert_matching_tables(self.shopify.get_orders(None, 2, False), self.mock_result_since)
        assert_matching_tables(
            self.shopify.get_orders(None, None, True), self.mock_result_completed
        )

    @requests_mock.Mocker()
    def test_get_query_url(self, m):
        self.assertEqual(
            self.shopify.get_query_url(None, None, "orders", True),
            f"https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders/"
            "count.json?limit=250&status=any",
        )
        self.assertEqual(
            self.shopify.get_query_url("2020-10-20", None, "orders", True),
            f"https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders/"
            "count.json?limit=250&status=any&created_at_min=2020-10-20T00:00:00&"
            "created_at_max=2020-10-21T00:00:00",
        )
        self.assertEqual(
            self.shopify.get_query_url(None, 2, "orders", True),
            f"https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders/"
            "count.json?limit=250&status=any&since_id=2",
        )
        self.assertEqual(
            self.shopify.get_query_url(None, None, "orders", False),
            f"https://{SUBDOMAIN}.myshopify.com/admin/api/{API_VERSION}/orders.json?"
            "limit=250&status=any",
        )

    @requests_mock.Mocker()
    def test_graphql(self, m):
        m.post(
            "https://{0}.myshopify.com/admin/api/{1}/graphql.json".format(SUBDOMAIN, API_VERSION),
            json=self.mock_graphql,
        )
        self.assertEqual(
            self.shopify.graphql(
                """
            {{
                orders(query: "financial_status:=paid", first: 100) {{
                    edges {{
                        node {{
                            id
                        }}
                    }}
                }}
            }}
        """
            ),
            self.mock_graphql["data"],
        )



import json

GET_ORGANIZATIONS_JSON = {
    "count": 38,
    "next": None,
    "previous": ("https://events.mobilizeamerica.io/api/v1/organizations?updated_since=1543644000"),
    "data": [
        {
            "id": 1251,
            "name": "Mike Blake for New York City",
            "slug": "mikefornyc",
            "is_coordinated": "True",
            "is_independent": "True",
            "is_primary_campaign": "False",
            "state": "",
            "district": "",
            "candidate_name": "",
            "race_type": "OTHER_LOCAL",
            "event_feed_url": "https://events.mobilizeamerica.io/mikefornyc/",
            "created_date": 1545885434,
            "modified_date": 1546132256,
        }
    ],
}

GET_EVENTS_JSON = {
    "count": 1,
    "next": None,
    "previous": None,
    "data": [
        {
            "id": 86738,
            "description": (
                "Join our team of volunteers and learn how to engage students in local "
                "high schools, communicate our mission, and register young voters."
            ),
            "timezone": "America/Chicago",
            "title": "Student Voter Initiative Training",
            "summary": "",
            "featured_image_url": (
                "https://mobilizeamerica.imgix.net/uploads/event/"
                "40667432145_6188839fe3_o_20190102224312253645.jpeg"
            ),
            "sponsor": {
                "id": 1076,
                "name": "Battleground Texas",
                "slug": "battlegroundtexas",
                "is_coordinated": True,
                "is_independent": False,
                "is_primary_campaign": False,
                "state": "",
                "district": "",
                "candidate_name": "",
                "race_type": None,
                "event_feed_url": "https://events.mobilizeamerica.io/battlegroundtexas/",
                "created_date": 1538590930,
                "modified_date": 1546468308,
            },
            "timeslots": [{"id": 526226, "start_date": 1547330400, "end_date": 1547335800}],
            "location": {
                "venue": "Harris County Democratic Party HQ",
                "address_lines": ["4619 Lyons Ave", ""],
                "locality": "Houston",
                "region": "TX",
                "postal_code": "77020",
                "location": {"latitude": 29.776446, "longitude": -95.323037},
                "congressional_district": "18",
                "state_leg_district": "142",
                "state_senate_district": None,
            },
            "event_type": "TRAINING",
            "created_date": 1546469706,
            "modified_date": 1547335800,
            "browser_url": ("https://events.mobilizeamerica.io/battlegroundtexas/event/86738/"),
            "high_priority": None,
            "contact": None,
            "visibility": "PUBLIC",
        }
    ],
}

GET_EVENTS_ORGANIZATION_JSON = json.loads(
    """
    {
        "count": 2,
        "next": null,
        "previous": null,
        "data": [
            {
            "approval_status": "APPROVED",
            "address_visibility": "PUBLIC",
            "location": {
                "venue": "Test",
                "address_lines": [
                "123 Test Road",
                ""
                ],
                "locality": "York",
                "region": "PA",
                "country": "US",
                "postal_code": "17404",
                "location": {
                "latitude": 40.0588876,
                "longitude": -76.7835604
                },
                "congressional_district": null,
                "state_leg_district": null,
                "state_senate_district": null
            },
            "timeslots": [
                {
                "id": 1,
                "start_date": 2,
                "end_date": 3,
                "instructions": ""
                },
                {
                "id": 2,
                "start_date": 3,
                "end_date": 4,
                "instructions": "Some detailed instructions for the second timeslot"
                }
            ],
            "title": "Test Event",
            "accessibility_status": null,
            "created_date": 1532629574,
            "created_by_volunteer_host": false,
            "instructions": "",
            "virtual_action_url": null,
            "summary": "",
            "sponsor": {
                "id": 1,
                "is_primary_campaign": false,
                "name": "Test Org",
                "is_independent": true,
                "candidate_name": "",
                "org_type": "OTHER",
                "created_date": 1513974036,
                "event_feed_url": "https://www.mobilize.us/test_org/",
                "state": "",
                "race_type": null,
                "logo_url": "https://mobilize-uploads-prod.s3.amazonaws.com/branding/test_logo.png",
                "is_coordinated": false,
                "district": "",
                "slug": "testorg",
                "modified_date": 1655222024
            },
            "featured_image_url": "",
            "contact": {
                "name": "Test",
                "email_address": "tester@test.org",
                "phone_number": "",
                "owner_user_id": 1234
            },
            "timezone": "America/New_York",
            "id": 7659,
            "description": "Test",
            "event_campaign": null,
            "high_priority": false,
            "accessibility_notes": null,
            "event_type": "PHONE_BANK",
            "browser_url": "https://www.mobilize.us/test_org/event/1/",
            "visibility": "PUBLIC",
            "is_virtual": false,
            "modified_date": 1601663981,
            "tags": []
            },
            {
            "approval_status": "APPROVED",
            "address_visibility": "PRIVATE",
            "location": {
                "locality": "Schenectady",
                "region": "NY",
                "country": "US",
                "postal_code": "12309",
                "congressional_district": "20",
                "state_leg_district": "110",
                "state_senate_district": "49",
                "address_lines": [
                "This event address is private. Sign up for more details",
                ""
                ],
                "venue": "This event address is private. Sign up for more details"
            },
            "timeslots": [
                {
                "id": 1,
                "start_date": 2,
                "end_date": 3,
                "instructions": ""
                },
                {
                "id": 2,
                "start_date": 3,
                "end_date": 4,
                "instructions": "Some detailed instructions for the second timeslot"
                }
            ],
            "title": "Test Phonebank",
            "accessibility_status": null,
            "created_date": 1537289907,
            "created_by_volunteer_host": false,
            "instructions": null,
            "virtual_action_url": null,
            "summary": "Help Us Call Testers!",
            "sponsor": {
                "id": 321,
                "is_primary_campaign": false,
                "name": "Test Org Two",
                "is_independent": false,
                "candidate_name": "",
                "org_type": "TEST_ORG",
                "created_date": 1537214527,
                "event_feed_url": "https://www.mobilize.us/test_org_two/",
                "state": "NY",
                "race_type": null,
                "logo_url": "https://amazonaws.com/test_org_two.jpg",
                "is_coordinated": true,
                "district": "",
                "slug": "testorgtwo",
                "modified_date": 1654183362
            },
            "featured_image_url": "https://mobilizeamerica.imgix.net/uploads/test.jpg",
            "contact": null,
            "timezone": "America/New_York",
            "id": 421,
            "description": "Join us to call people and do an automated test!",
            "event_campaign": null,
            "high_priority": false,
            "accessibility_notes": null,
            "event_type": "PHONE_BANK",
            "browser_url": "https://www.mobilize.us/test_org_two/event/2/",
            "visibility": "PUBLIC",
            "is_virtual": false,
            "modified_date": 1601665649,
            "tags": []
            }
        ],
        "metadata": {
            "url_name": "public_organization_events",
            "build_commit": "abcd",
            "page_title": null
        }
    }"""
)

GET_EVENTS_DELETED_JSON = {
    "count": 2,
    "next": None,
    "previous": None,
    "data": [
        {"id": 86765, "deleted_date": 1546705971},
        {"id": 86782, "deleted_date": 1546912779},
    ],
}



# These are reserved words by Redshift and cannot be used as column names.
RESERVED_WORDS = [
    "AES128",
    "AES256",
    "ALL",
    "ALLOWOVERWRITE",
    "ANALYSE",
    "ANALYZE",
    "AND",
    "ANY",
    "ARRAY",
    "AS",
    "ASC",
    "AUTHORIZATION",
    "BACKUP",
    "BETWEEN",
    "BINARY",
    "BLANKSASNULL",
    "BOTH",
    "BYTEDICT",
    "BZIP2",
    "CASE",
    "CAST",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "CONSTRAINT",
    "CREATE",
    "CREDENTIALS",
    "CROSS",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "CURRENT_USER_ID",
    "DEFAULT",
    "DEFERRABLE",
    "DEFLATE",
    "DEFRAG",
    "DELTA",
    "DELTA32K",
    "DESC",
    "DISABLE",
    "DISTINCT",
    "DO",
    "ELSE",
    "EMPTYASNULL",
    "ENABLE",
    "ENCODE",
    "ENCRYPT",
    "ENCRYPTION",
    "END",
    "EXCEPT",
    "EXPLICIT",
    "FALSE",
    "FOR",
    "FOREIGN",
    "FREEZE",
    "FROM",
    "FULL",
    "GLOBALDICT256",
    "GLOBALDICT64K",
    "GRANT",
    "GROUP",
    "GZIP",
    "HAVING",
    "IDENTITY",
    "IGNORE",
    "ILIKE",
    "IN",
    "INITIALLY",
    "INNER",
    "INTERSECT",
    "INTO",
    "IS",
    "ISNULL",
    "JOIN",
    "LEADING",
    "LEFT",
    "LIKE",
    "LIMIT",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "LUN",
    "LUNS",
    "LZO",
    "LZOP",
    "MINUS",
    "MOSTLY13",
    "MOSTLY32",
    "MOSTLY8",
    "NATURAL",
    "NEW",
    "NOT",
    "NOTNULL",
    "NULL",
    "NULLS",
    "OFF",
    "OFFLINE",
    "OFFSET",
    "OLD",
    "ON",
    "ONLY",
    "OPEN",
    "OR",
    "ORDER",
    "OUTER",
    "OVERLAPS",
    "PARALLEL",
    "PARTITION",
    "PERCENT",
    "PERMISSIONS",
    "PLACING",
    "PRIMARY",
    "RAW",
    "READRATIO",
    "RECOVER",
    "REFERENCES",
    "RESPECT",
    "REJECTLOG",
    "RESORT",
    "RESTORE",
    "RIGHT",
    "SELECT",
    "SESSION_USER",
    "SIMILAR",
    "SOME",
    "SYSDATE",
    "SYSTEM",
    "TABLE",
    "TAG",
    "TDES",
    "TEXT255",
    "TEXT32K",
    "THEN",
    "TIMESTAMP",
    "TO",
    "TOP",
    "TRAILING",
    "TRUE",
    "TRUNCATECOLUMNS",
    "UNION",
    "UNIQUE",
    "USER",
    "USING",
    "VERBOSE",
    "WALLET",
    "WHEN",
    "WHERE",
    "WITH",
    "WITHOUT",
]

VARCHAR = "varchar"
FLOAT = "float"
BOOL = "bool"

# The following values are the minimum and maximum values for MySQL int
# types. https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
SMALLINT = "smallint"
SMALLINT_MIN = -32768
SMALLINT_MAX = 32767

MEDIUMINT = "mediumint"
MEDIUMINT_MIN = -8388608
MEDIUMINT_MAX = 8388607

INT = "int"
INT_MIN = -2147483648
INT_MAX = 2147483647

BIGINT = "bigint"

INT_TYPES = [SMALLINT, MEDIUMINT, INT, BIGINT]

COL_NAME_MAX_LEN = 100

IS_CASE_SENSITIVE = False

REPLACE_CHARS = {" ": "_"}



import tempfile
import unittest

from parsons import Table
from parsons.databases.sqlite import Sqlite
from test.utils import assert_matching_tables


class TestSqlite(unittest.TestCase):
    def setUp(self):
        temp_db = tempfile.mkstemp(suffix=".db")[1]
        self.sqlite = Sqlite(temp_db)

        self.tbl = Table([["ID", "Name"], [1, "Jim"], [2, "John"], [3, "Sarah"]])

    def test_copy(self) -> None:
        self.sqlite.copy(self.tbl, "tbl1", if_exists="drop")
        tbl1_read = self.sqlite.query("select * from tbl1")
        assert_matching_tables(self.tbl, tbl1_read)

    def test_copy_no_cli(self) -> None:
        self.sqlite.copy(self.tbl, "tbl1", if_exists="drop", force_python_sdk=True)
        tbl1_read = self.sqlite.query("select * from tbl1")
        assert_matching_tables(self.tbl, tbl1_read)

    def test_copy_append(self) -> None:
        self.sqlite.copy(self.tbl, "tbl1", if_exists="drop")
        self.sqlite.copy(self.tbl, "tbl1", if_exists="append")
        row_count = self.sqlite.query("select count(*) as count from tbl1")
        assert row_count[0]["count"] == 6

    def test_copy_fail(self) -> None:
        self.sqlite.copy(self.tbl, "tbl1", if_exists="drop")
        try:
            # This line should raise a ValueError
            self.sqlite.copy(self.tbl, "tbl1", if_exists="fail")
            # so this line should not be executed
            raise AssertionError("should have failed")
        except ValueError:
            pass

    def test_copy_truncate(self) -> None:
        self.sqlite.copy(self.tbl, "tbl1", if_exists="drop")
        self.sqlite.copy(self.tbl, "tbl1", if_exists="truncate")
        row_count = self.sqlite.query("select count(*) as count from tbl1")
        assert row_count[0]["count"] == 3


if __name__ == "__main__":
    unittest.main()



import gzip
import json
import logging
import time

import requests

from parsons.etl.table import Table
from parsons.utilities import check_env

logger = logging.getLogger(__name__)


class Auth0(object):
    """
    Instantiate the Auth0 class

    `Args:`
        client_id: str
            The Auth0 client ID. Not required if ``AUTH0_CLIENT_ID`` env variable set.
        client_secret: str
            The Auth0 client secret. Not required if ``AUTH0_CLIENT_SECRET`` env variable set.
        domain: str
            The Auth0 domain. Not required if ``AUTH0_DOMAIN`` env variable set.
    `Returns:`
        Auth0 Class
    """

    def __init__(self, client_id=None, client_secret=None, domain=None):
        self.base_url = f"https://{check_env.check('AUTH0_DOMAIN', domain)}"
        access_token = (
            requests.post(
                f"{self.base_url}/oauth/token",
                data={
                    "grant_type": "client_credentials",  # OAuth 2.0 flow to use
                    "client_id": check_env.check("AUTH0_CLIENT_ID", client_id),
                    "client_secret": check_env.check("AUTH0_CLIENT_SECRET", client_secret),
                    "audience": f"{self.base_url}/api/v2/",
                },
            )
            .json()
            .get("access_token")
        )
        self.headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json",
        }

    def delete_user(self, id):
        """
        Delete Auth0 user.

        `Args:`
            id: str
                The user ID of the record to delete.
        `Returns:`
            int
        """
        return requests.delete(
            f"{self.base_url}/api/v2/users/{id}", headers=self.headers
        ).status_code

    def get_users_by_email(self, email):
        """
        Get Auth0 users by email.

        `Args:`
            email: str
                The user email of the record to get.
        `Returns:`
            Table Class
        """
        url = f"{self.base_url}/api/v2/users-by-email"
        val = requests.get(url, headers=self.headers, params={"email": email})
        if val.status_code == 429:
            raise requests.exceptions.ConnectionError(val.json()["message"])
        return Table(val.json())

    def upsert_user(
        self,
        email,
        username=None,
        given_name=None,
        family_name=None,
        app_metadata=None,
        user_metadata=None,
        connection="Username-Password-Authentication",
    ):
        """
        Upsert Auth0 users by email.

        `Args:`
            email: str
                The user email of the record to get.
            username: optional str
                Username to set for user
            given_name: optional str
                Given to set for user
            family_name: optional str
                Family name to set for user
            app_metadata: optional dict
                App metadata to set for user
            user_metadata: optional dict
                User metadata to set for user
        `Returns:`
            Requests Response object
        """

        if user_metadata is None:
            user_metadata = {}
        if app_metadata is None:
            app_metadata = {}
        obj = {
            "email": email.lower(),
            "username": username,
            "connection": connection,
            "app_metadata": app_metadata,
            "blocked": False,
            "user_metadata": user_metadata,
        }
        if given_name is not None:
            obj["given_name"] = given_name
        if family_name is not None:
            obj["family_name"] = family_name
        payload = json.dumps(obj)

        existing = self.get_users_by_email(email.lower())
        if existing.num_rows > 0:
            a0id = existing[0]["user_id"]
            ret = requests.patch(
                f"{self.base_url}/api/v2/users/{a0id}",
                headers=self.headers,
                data=payload,
            )
        else:
            ret = requests.post(f"{self.base_url}/api/v2/users", headers=self.headers, data=payload)
        if ret.status_code != 200:
            raise ValueError(f"Invalid response {ret.json()}")
        return ret

    def block_user(self, user_id, connection="Username-Password-Authentication"):
        """
        Blocks Auth0 users by email - setting the "blocked" attribute on Auth0's API.

        `Args:`
            user_id: str
                Auth0 user id
            connection: optional str
                Name of auth0 connection (default to Username-Password-Authentication)
        `Returns:`
            Requests Response object
        """
        payload = json.dumps({"connection": connection, "blocked": True})
        ret = requests.patch(
            f"{self.base_url}/api/v2/users/{user_id}",
            headers=self.headers,
            data=payload,
        )
        if ret.status_code != 200:
            raise ValueError(f"Invalid response {ret.json()}")
        return ret

    def retrieve_all_users(self, connection="Username-Password-Authentication"):
        """
        Retrieves all Auth0 users using the batch jobs endpoint.

        `Args:`
            connection: optional str
                Name of auth0 connection (default to Username-Password-Authentication)
        `Returns:`
            Requests Response object
        """
        connection_id = self.get_connection_id(connection)
        url = f"{self.base_url}/api/v2/jobs/users-exports"

        headers = self.headers

        fields = [
            {"name": n} for n in ["user_id", "username", "email", "user_metadata", "app_metadata"]
        ]
        # Start the users-export job
        response = requests.post(
            url,
            headers=headers,
            json={"connection_id": connection_id, "format": "json", "fields": fields},
        )
        job_id = response.json().get("id")

        if job_id:
            # Check job status until complete
            while True:
                status_response = requests.get(
                    f"{self.base_url}/api/v2/jobs/{job_id}", headers=headers
                )
                status_data = status_response.json()
                if status_response.status_code == 429:
                    time.sleep(10)

                elif status_response.status_code != 200:
                    break
                elif status_data.get("status") == "completed":
                    download_url = status_data.get("location")
                    break
                elif status_data.get("status") == "failed":
                    logger.error("Retrieve members job failed to complete.")
                    return None

            # Download the users-export file
            users_response = requests.get(download_url)

            decompressed_data = gzip.decompress(users_response.content).decode("utf-8")
            users_data = []
            for d in decompressed_data.split("\n"):
                if d:
                    users_data.append(json.loads(d))

            return Table(users_data)

        logger.error("Retrieve members job creation failed")
        return None

    def get_connection_id(self, connection_name):
        """
        Retrieves an Auth0 connection_id corresponding to a specific connection name

        `Args:`
            connection_name: str
                Name of auth0 connection
        `Returns:`
            Connection ID string
        """
        url = f"{self.base_url}/api/v2/connections"

        response = requests.get(url, headers=self.headers)
        connections = response.json()

        for connection in connections:
            if connection["name"] == connection_name:
                return connection["id"]

        return None



class Locations:
    """A class for getting, creating, and editing PDI locations"""

    def __init__(self):
        self.locations_url = self.base_url + "/locations"

        super().__init__()

    def get_locations(self, limit=None):
        """Get a list of PDI Locations

        `Args:`
            limit: int
                The max number of locations to return

        `Returns:`
            parsons.Table
                A Parsons table containing all requested location data.
        """

        return self._request(self.locations_url, limit=limit)

    def create_location(self, address: str, name: str):
        """Create a new PDI address
        `Args:`
            address: str
               A full address including street number, city, state, and zip.
            name: str
                The name of the location. E.g. "The Overlook Hotel"
        `Returns:`
            dict
                Response from PDI in dictionary object
        """

        payload = {"locationName": name, "locationAddress": address}
        return self._request(self.locations_url, req_type="POST", post_data=payload)

    def get_location(self, id: str):
        return self._request(f"{self.locations_url}/{id}")

    def update_location(self, id: str, location_name: str, address: str):
        payload = {"locationName": location_name, "locationAddress": address}
        res = self._request(f"{self.locations_url}/{id}", req_type="PUT", post_data=payload)
        if res["code"] == 201:
            return True
